import random
from dataclasses import dataclass
from typing import Dict, Mapping, Optional

import numpy as np
import numpy.typing as npt

from entity_gym.dataclass_utils import extract_features, obs_space_from_dataclasses
from entity_gym.env import (
    Action,
    ActionSpace,
    CategoricalAction,
    CategoricalActionMask,
    CategoricalActionSpace,
    Environment,
    Observation,
    ObsSpace,
)


@dataclass
class Player:
    pass


@dataclass
class Bean:
    pass


class Count(Environment):
    """
    There are between 0 and 10 "Bean" entities.
    The "Player" entity gets 1 reward for counting the correct number of beans and 0 otherwise.

    This environment also randomly masks off some of the incorrect answers.

    Masking by default allows all actions, which is equivalent to disabling masking.
    """

    def __init__(self, masked_choices: int = 10):
        assert (
            masked_choices >= 1 and masked_choices <= 10
        ), "masked_choices must be between 1 and 10"
        self.masked_choices = masked_choices

    def obs_space(self) -> ObsSpace:
        return obs_space_from_dataclasses(Player, Bean)

    def action_space(self) -> Dict[str, ActionSpace]:
        return {
            "count": CategoricalActionSpace(
                ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]
            )
        }

    def reset_filter(self, obs_space: ObsSpace) -> Observation:
        self.count = random.randint(0, self.masked_choices - 1)
        possible_counts = {
            self.count,
            *random.sample(
                range(0, self.masked_choices), random.randint(0, self.masked_choices)
            ),
        }
        mask = np.zeros((1, 10), dtype=np.bool_)
        mask[:, list(possible_counts)] = True
        return self.observe(obs_space, mask)

    def reset(self) -> Observation:
        return self.reset_filter(self.obs_space())

    def act_filter(
        self, action: Mapping[str, Action], obs_filter: ObsSpace
    ) -> Observation:
        reward = 0.0
        assert len(action) == 1
        a = action["count"]
        assert isinstance(a, CategoricalAction)
        assert len(a.indices) == 1
        choice = a.indices[0]
        if choice == self.count:
            reward = 1.0
        return self.observe(obs_filter, None, done=True, reward=reward)

    def act(self, actions: Mapping[str, Action]) -> Observation:
        return self.act_filter(
            actions,
            self.obs_space(),
        )

    def observe(
        self,
        obs_filter: ObsSpace,
        mask: Optional[npt.NDArray[np.bool_]],
        done: bool = False,
        reward: float = 0.0,
    ) -> Observation:
        return Observation(
            features=extract_features(
                {
                    "Player": [Player()],
                    "Bean": [Bean()] * self.count,
                },
                obs_filter,
            ),
            actions={
                "count": CategoricalActionMask(actor_ids=["Player"], mask=mask),
            },
            ids={
                "Player": ["Player"],
                "Bean": [f"Bean{i}" for i in range(1, self.count + 1)],
            },
            reward=reward,
            done=done,
        )



import random
from dataclasses import dataclass, field
from typing import Dict, List, Mapping, Tuple

import numpy as np

from entity_gym.dataclass_utils import extract_features, obs_space_from_dataclasses
from entity_gym.env import (
    Action,
    ActionSpace,
    CategoricalAction,
    CategoricalActionMask,
    CategoricalActionSpace,
    Environment,
    Observation,
    ObsSpace,
)


@dataclass
class Vehicle:
    x_pos: float = 0.0
    y_pos: float = 0.0
    direction: float = 0.0
    step: int = 0


@dataclass
class Target:
    x_pos: float = 0.0
    y_pos: float = 0.0


@dataclass
class Mine:
    x_pos: float = 0.0
    y_pos: float = 0.0


@dataclass
class Minefield(Environment):
    """
    Task with a Vehicle entity that has to reach a target point, receiving a reward of 1.
    If the vehicle collides with any of the randomly placed mines, the episode ends without reward.
    The available actions either turn the vehicle left, right, or go straight.
    """

    vehicle: Vehicle = field(default_factory=Vehicle)
    target: Target = field(default_factory=Target)
    mine: Mine = field(default_factory=Mine)
    max_mines: int = 10
    max_steps: int = 200
    translate: bool = False
    width: float = 200.0

    def obs_space(self) -> ObsSpace:
        return obs_space_from_dataclasses(Vehicle, Mine, Target)

    def action_space(self) -> Dict[str, ActionSpace]:
        return {
            "move": CategoricalActionSpace(
                ["turn left", "move forward", "turn right"],
            )
        }

    def reset_filter(self, obs_space: ObsSpace) -> Observation:
        def randpos() -> Tuple[float, float]:
            return (
                random.uniform(-self.width / 2, self.width / 2),
                random.uniform(-self.width / 2, self.width / 2),
            )

        self.vehicle.x_pos, self.vehicle.y_pos = randpos()
        self.target.x_pos, self.target.y_pos = randpos()
        mines: List[Mine] = []
        for _ in range(self.max_mines):
            x, y = randpos()
            # Check that the mine is not too close to the vehicle, target, or any other mine
            pos = [(m.x_pos, m.y_pos) for m in mines] + [
                (self.vehicle.x_pos, self.vehicle.y_pos),
                (self.target.x_pos, self.target.y_pos),
            ]
            if any(map(lambda p: (x - p[0]) ** 2 + (y - p[1]) ** 2 < 15 * 15, pos)):
                continue
            mines.append(Mine(x, y))
        self.vehicle.direction = random.uniform(0, 2 * np.pi)
        self.step = 0
        self.mines = mines
        return self.observe(obs_space)

    def reset(self) -> Observation:
        return self.reset_filter(self.obs_space())

    def act_filter(
        self, action: Mapping[str, Action], obs_filter: ObsSpace
    ) -> Observation:
        for action_name, a in action.items():
            assert isinstance(a, CategoricalAction)
            if action_name == "move":
                move = a.indices[0]
                if move == 0:
                    self.vehicle.direction -= np.pi / 8
                elif move == 1:
                    self.vehicle.x_pos += 3 * np.cos(self.vehicle.direction)
                    self.vehicle.y_pos += 3 * np.sin(self.vehicle.direction)
                elif move == 2:
                    self.vehicle.direction += np.pi / 8
                else:
                    raise ValueError(
                        f"Invalid action {move} for action space {action_name}"
                    )
                self.vehicle.direction %= 2 * np.pi
            else:
                raise ValueError(f"Unknown action type {action_name}")

        self.step += 1
        self.vehicle.step = self.step

        return self.observe(obs_filter)

    def act(self, actions: Mapping[str, Action]) -> Observation:
        return self.act_filter(
            actions,
            self.obs_space(),
        )

    def observe(self, obs_filter: ObsSpace, done: bool = False) -> Observation:
        if (self.target.x_pos - self.vehicle.x_pos) ** 2 + (
            self.target.y_pos - self.vehicle.y_pos
        ) ** 2 < 5 * 5:
            done = True
            reward = 1
        elif (
            any(
                map(
                    lambda m: (self.vehicle.x_pos - m.x_pos) ** 2
                    + (self.vehicle.y_pos - m.y_pos) ** 2
                    < 5 * 5,
                    self.mines,
                )
            )
            or self.step >= self.max_steps
        ):
            done = True
            reward = 0
        else:
            done = False
            reward = 0

        if self.translate:
            ox = self.vehicle.x_pos
            oy = self.vehicle.y_pos
        else:
            ox = oy = 0
        return Observation(
            features=extract_features(
                {
                    "Mine": [Mine(m.x_pos - ox, m.y_pos - oy) for m in self.mines],
                    "Vehicle": [self.vehicle],
                    "Target": [Target(self.target.x_pos - ox, self.target.y_pos - oy)],
                },
                obs_filter,
            ),
            actions={
                "move": CategoricalActionMask(actor_types=["Vehicle"]),
            },
            ids={"Vehicle": ["Vehicle"]},
            reward=reward,
            done=done,
        )



"""
The environment module defines the core interfaces that make up an `Environment <#entity_gym.env.Environment>`_.

Actions
-------

Actions are how agents interact with the environment.
There are three parts to every action:

* :code:`ActionSpace` defines the shape of the action. For example, a categorical action space consisting of the 4 discrete choices "up", "down", "left", and "right".
* :code:`ActionMask` is used to further constrain the available actions on a specific timestep. For example, only "up" and "down" may be available some timestep.
* :code:`Action` represent the actual action that is chosen by an agent. For example, the "down" action may have been chosen.

There are currently three different action spaces:

* `GlobalCategoricalActionSpace <#entity_gym.env.GlobalCategoricalActionSpace>`_ allows the agent to choose a single option from a discrete set of actions.
* `CategoricalActionSpace <#entity_gym.env.CategoricalActionSpace>`_ allows multiple entities to choose a single option from a discrete set of actions.
* `SelectEntityActionSpace <#entity_gym.env.SelectEntityActionSpace>`_ allows multiple entities to choose another entity.

Observations
------------

Observations are how agents receive information from the environment.
Each `Environment <#entity_gym.env.Environment>`_ must define an `ObsSpace <#entity_gym.env.ObsSpace>`_, which specifies the shape of the observations returned by this environment.
On each timestep, the environment returns an `Observation <#entity_gym.env.Observation>`_ object, which contains all the entities and features that are visible to the agent.
"""
from .action import *
from .add_metrics_wrapper import AddMetricsWrapper
from .env_list import *
from .environment import *
from .parallel_env_list import *
from .validator import ValidatingEnv
from .vec_env import *

__all__ = [
    "Environment",
    # Observation
    "ObsSpace",
    "Entity",
    "Observation",
    "EntityName",
    "ActionName",
    "EntityID",
    # Action
    "Action",
    "ActionSpace",
    "ActionMask",
    "CategoricalActionSpace",
    "CategoricalAction",
    "CategoricalActionMask",
    "GlobalCategoricalActionSpace",
    "GlobalCategoricalAction",
    "GlobalCategoricalActionMask",
    "SelectEntityAction",
    "SelectEntityActionSpace",
    "SelectEntityActionMask",
    # VecEnv
    "VecEnv",
    "EnvList",
    "ParallelEnvList",
    "VecActionMask",
    "VecObs",
    "VecCategoricalActionMask",
    "VecSelectEntityActionMask",
    # Wrappers
    "ValidatingEnv",
    "AddMetricsWrapper",
]



from dataclasses import dataclass
from typing import List, Optional, Sequence, Union

import numpy as np
import numpy.typing as npt

from .common import EntityID, EntityName


@dataclass
class CategoricalActionSpace:
    """
    Defines a discrete set of actions that can be taken by multiple entities.
    """

    index_to_label: List[str]
    """list of human-readable labels for each action"""

    def __len__(self) -> int:
        return len(self.index_to_label)


@dataclass
class GlobalCategoricalActionSpace:
    """
    Defines a discrete set of actions that can be taken on each timestep.

    For example, the following actions space allows the agent to choose between four actions "up", "down", "left", and "right":

    .. code-block:: python

        GlobalCategoricalActionSpace(["up", "down", "left", "right"])
    """

    index_to_label: List[str]
    """list of human-readable labels for each action"""

    def __len__(self) -> int:
        return len(self.index_to_label)


@dataclass
class SelectEntityActionSpace:
    """
    Allows multiple entities to each select another entity.
    """


ActionSpace = Union[
    CategoricalActionSpace, SelectEntityActionSpace, GlobalCategoricalActionSpace
]


@dataclass
class CategoricalActionMask:
    """
    Action mask for categorical action that specifies which agents can perform the action,
    and includes a dense mask that further constraints the choices available to each agent.
    """

    actor_ids: Optional[Sequence[EntityID]] = None
    """
    The ids of the entities that can perform the action.
    If ``None``, all entities can perform the action.
    Mutually exclusive with ``actor_types``.
    """

    actor_types: Optional[Sequence[EntityName]] = None
    """
    The types of the entities that can perform the action.
    If ``None``, all entities can perform the action.
    Mutually exclusive with ``actor_ids``.
    """

    mask: Union[Sequence[Sequence[bool]], np.ndarray, None] = None
    """
    A boolean array of shape ``(len(actor_ids), len(choices))`` that prevents specific actions from being available to certain entities.
    If ``mask[i, j]`` is ``True``, then the entity with id ``actor_ids[i]`` can perform action ``j``.
    """

    def __post_init__(self) -> None:
        assert (
            self.actor_ids is None or self.actor_types is None
        ), "Only one of actor_ids or actor_types can be specified"


@dataclass
class GlobalCategoricalActionMask:
    """
    Action mask for global categorical action.
    """

    mask: Union[Sequence[Sequence[bool]], np.ndarray, None] = None
    """
    An optional boolean array of shape (len(choices),). If mask[i] is True, then
    action choice i can be performed.
    """


@dataclass
class SelectEntityActionMask:
    """
    Action mask for select entity action that specifies which agents can perform the action,
    and includes a dense mask that further constraints what other entities can be selected by
    each actor.
    """

    actor_ids: Optional[Sequence[EntityID]] = None
    """
    The ids of the entities that can perform the action.
    If None, all entities can perform the action.
    """

    actor_types: Optional[Sequence[EntityName]] = None
    """
    The types of the entities that can perform the action.
    If None, all entities can perform the action.
    """

    actee_types: Optional[Sequence[EntityName]] = None
    """
    The types of entities that can be selected by each actor.
    If None, all entities types can be selected by each actor.
    """

    actee_ids: Optional[Sequence[EntityID]] = None
    """
    The ids of the entities of each type that can be selected by each actor.
    If None, all entities can be selected by each actor.
    """

    mask: Optional[npt.NDArray[np.bool_]] = None
    """
    An boolean array of shape (len(actor_ids), len(actee_ids)). If mask[i, j] is True, then
    the agent with id actor_ids[i] can select entity with id actee_ids[j].
    (NOT CURRENTLY IMPLEMENTED)
    """

    def __post_init__(self) -> None:
        assert (
            self.actor_ids is None or self.actor_types is None
        ), "Only one of actor_ids or actor_types can be specified"
        assert (
            self.actee_types is None or self.actee_ids is None
        ), "Either actee_entity_types or actees can be specified, but not both."


ActionMask = Union[
    CategoricalActionMask, SelectEntityActionMask, GlobalCategoricalActionMask
]


@dataclass
class CategoricalAction:
    """
    Outcome of a categorical action.
    """

    actors: Sequence[EntityID]
    """the ids of the entities that chose the actions"""

    indices: npt.NDArray[np.int64]
    """the indices of the actions that were chosen"""

    index_to_label: List[str]
    """mapping from action indices to human readable labels"""

    probs: Optional[npt.NDArray[np.float32]] = None
    """the probablity assigned to each action by each agent"""

    @property
    def labels(self) -> List[str]:
        """the human readable labels of the actions that were performed"""
        return [self.index_to_label[i] for i in self.indices]


@dataclass
class SelectEntityAction:
    """
    Outcome of a select entity action.
    """

    actors: Sequence[EntityID]
    """the ids of the entities that chose the action"""
    actees: Sequence[EntityID]
    """the ids of the entities that were selected by the actors"""
    probs: Optional[npt.NDArray[np.float32]] = None
    """the probablity assigned to each selection by each agent"""


@dataclass
class GlobalCategoricalAction:
    """Outcome of a global categorical action."""

    index: int
    """the index of the action that was chosen"""
    label: str
    """the human readable label of the action that was chosen"""
    probs: Optional[npt.NDArray[np.float32]] = None
    """the probablity assigned to the action by each agent"""


Action = Union[CategoricalAction, SelectEntityAction, GlobalCategoricalAction]



from dataclasses import is_dataclass
from typing import Any, Dict, List, Type

import numpy as np

from .env import Entity, ObsSpace


def obs_space_from_dataclasses(*dss: Type) -> ObsSpace:
    entities = {}
    for ds in dss:
        if not is_dataclass(ds):
            raise ValueError(f"{ds} is not a dataclass")
        # TODO: check field types are valid
        entities[ds.__name__] = Entity(
            features=list(
                key for key in ds.__dataclass_fields__.keys() if not key.startswith("_")
            ),
        )
    return ObsSpace(entities=entities)


def extract_features(
    entities: Dict[str, List[Any]], obs_filter: ObsSpace
) -> Dict[str, np.ndarray]:
    selectors = {}
    for entity_name, entity in obs_filter.entities.items():
        selectors[entity_name] = np.array(
            [[getattr(e, f) for f in entity.features] for e in entities[entity_name]],
            dtype=np.float32,
        ).reshape(len(entities[entity_name]), len(entity.features))
    return selectors



from typing import Any, Dict, Mapping

import numpy as np
import numpy.typing as npt

from entity_gym.env.environment import (
    Action,
    ActionName,
    ActionSpace,
    CategoricalActionMask,
    CategoricalActionSpace,
    Environment,
    Observation,
    ObsSpace,
    SelectEntityActionMask,
    SelectEntityActionSpace,
)


class ValidatingEnv(Environment):
    def __init__(self, env: Environment) -> None:
        self.env = env
        self._obs_space = env.obs_space()
        self._action_space = env.action_space()

    def act(self, actions: Mapping[ActionName, Action]) -> Observation:
        obs = self.env.act(actions)
        try:
            self._validate(obs)
        except AssertionError as e:
            print(f"Invalid observation:\n{e}")
            raise e
        return obs

    def reset(self) -> Observation:
        obs = self.env.reset()
        try:
            self._validate(obs)
        except AssertionError as e:
            print(f"Invalid observation:\n{e}")
            raise e
        return obs

    def render(self, **kwargs: Any) -> npt.NDArray[np.uint8]:
        return self.env.render(**kwargs)

    def obs_space(self) -> ObsSpace:
        return self._obs_space

    def action_space(self) -> Dict[str, ActionSpace]:
        return self._action_space

    def _validate(self, obs: Observation) -> None:
        assert isinstance(
            obs, Observation
        ), f"Observation has invalid type: {type(obs)}"

        # Validate features
        for entity_type, entity_features in obs.features.items():
            assert (
                entity_type in self._obs_space.entities
            ), f"Features contain entity of type '{entity_type}' which is not in observation space: {list(self._obs_space.entities.keys())}"
            if isinstance(entity_features, np.ndarray):
                assert (
                    entity_features.dtype == np.float32
                ), f"Features of entity of type '{entity_type}' have invalid dtype: {entity_features.dtype}. Expected: {np.float32}"
                shape = entity_features.shape
                assert len(shape) == 2 and shape[1] == len(
                    self._obs_space.entities[entity_type].features
                ), f"Features of entity of type '{entity_type}' have invalid shape: {shape}. Expected: (n, {len(self._obs_space.entities[entity_type].features)})"
            else:
                for i, entity in enumerate(entity_features):
                    assert len(entity) == len(
                        self._obs_space.entities[entity_type].features
                    ), f"Features of {i}-th entity of type '{entity_type}' have invalid length: {len(entity)}. Expected: {len(self._obs_space.entities[entity_type].features)}"

            if entity_type in obs.ids:
                assert len(obs.ids[entity_type]) == len(
                    entity_features
                ), f"Length of ids of entity of type '{entity_type}' does not match length of features: {len(obs.ids[entity_type])} != {len(entity_features)}"

        # Validate global features
        if len(obs.global_features) != len(self._obs_space.global_features):
            raise AssertionError(
                f"Length of global features does not match length of global features in observation space: {len(obs.global_features)} != {len(self._obs_space.global_features)}"
            )

        # Validate ids
        previous_ids = set()
        for entity_type, entity_ids in obs.ids.items():
            assert (
                entity_type in self._obs_space.entities
            ), f"IDs contain entity of type '{entity_type}' which is not in observation space: {list(self._obs_space.entities.keys())}"
            for id in entity_ids:
                assert id not in previous_ids, f"Observation has duplicate id '{id}'"
                previous_ids.add(id)

        # Validate actions
        ids = obs.id_to_index(self._obs_space)
        for action_type, action_mask in obs.actions.items():
            assert (
                action_type in self._action_space
            ), f"Actions contain action of type '{action_type}' which is not in action space: {list(self._action_space.keys())}"
            space = self._action_space[action_type]
            if isinstance(space, CategoricalActionSpace):
                assert isinstance(
                    action_mask, CategoricalActionMask
                ), f"Action of type '{action_type}' has invalid type: {type(action_mask)}. Expected: CategoricalActionMask"
                if action_mask.actor_ids is not None:
                    for id in action_mask.actor_ids:
                        assert (
                            id in ids
                        ), f"Action of type '{action_type}' contains invalid actor id {id} which is not in ids: {obs.ids}"
                if action_mask.actor_types is not None:
                    for actor_type in action_mask.actor_types:
                        assert (
                            actor_type in obs.ids
                        ), f"Action of type '{action_type}' contains invalid actor type {actor_type} which is not in ids: {obs.ids.keys()}"
                mask = action_mask.mask
                if isinstance(mask, np.ndarray):
                    assert (
                        mask.dtype == np.bool_
                    ), f"Action of type '{action_type}' has invalid dtype: {mask.dtype}. Expected: {np.bool_}"
                    actor_indices = obs._actor_indices(action_type, self._obs_space)
                    shape = mask.shape
                    if shape[0] != 0:
                        assert shape == (
                            len(actor_indices),
                            len(space.index_to_label),
                        ), f"Action of type '{action_type}' has invalid shape: {shape}. Expected: ({len(actor_indices), len(space.index_to_label)})"
                        unmasked_count = mask.sum(axis=1)
                        for i in range(len(unmasked_count)):
                            assert (
                                unmasked_count[i] > 0
                            ), f"Action of type '{action_type}' contains invalid mask for {i}-th actor: {mask[i]}. Expected at least one possible action"
                elif mask is not None:
                    assert len(mask) == len(
                        actor_indices
                    ), f"Action of type '{action_type}' has invalid length: {len(mask)}. Expected: {len(actor_indices)}"
                    for i in range(len(mask)):
                        assert len(mask[i]) == len(
                            space.index_to_label
                        ), f"Action of type '{action_type}' has invalid length of mask for {i}-th actor: {len(mask[i])}. Expected: {len(space.index_to_label)}"
                        assert any(
                            mask[i]
                        ), f"Action of type '{action_type}' contains invalid mask for {i}-th actor: {mask[i]}. Expected at least one possible action"

            elif isinstance(self._action_space[action_type], SelectEntityActionSpace):
                assert isinstance(
                    action_mask, SelectEntityActionMask
                ), f"Action of type '{action_type}' has invalid type: {type(action_mask)}. Expected: SelectEntityActionMask"



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Dict, List, Mapping, Optional, Sequence, Tuple, Union

import numpy as np
import numpy.typing as npt

from .action import *
from .common import ActionName, EntityID, EntityName, Features


@dataclass
class Entity:
    """Defines the set of features for an entity type."""

    features: List[str]


@dataclass
class ObsSpace:
    """
    Defines what features can be observed by an agent.
    """

    global_features: List[str] = field(default_factory=list)
    """fixed size list of features that are observable on each timestep"""
    entities: Dict[EntityName, Entity] = field(default_factory=dict)
    """
    Defines the types of entities that can be observed.
    On a given timestep, an ``Observation`` may contain multiple entities of each type.
    """


class Observation:
    """
    Observation returned by the environment on one timestep.

    :param features: Maps each entity type to a list of features for the entities of that type.
    :param actions: Maps each action type to an ActionMask specifying which entities can perform
        the action.
    :param reward: Reward received on this timestep.
    :param done: Whether the episode has ended.
    :param ids: Maps each entity type to a list of entity ids for the entities of that type.
    :param visible: Optional mask for each entity type that prevents the policy but not the
        value function from observing certain entities.
    """

    global_features: Union[npt.NDArray[np.float32], Sequence[float]]
    features: Mapping[EntityName, Features]
    actions: Mapping[ActionName, ActionMask]
    done: bool
    reward: float
    ids: Mapping[EntityName, Sequence[EntityID]] = field(default_factory=dict)
    visible: Mapping[EntityName, Union[npt.NDArray[np.bool_], Sequence[bool]]] = field(
        default_factory=dict
    )
    metrics: Dict[str, float] = field(default_factory=dict)

    def __init__(
        self,
        *,
        done: bool,
        reward: float,
        visible: Optional[
            Mapping[EntityName, Union[npt.NDArray[np.bool_], Sequence[bool]]]
        ] = None,
        entities: Optional[
            Mapping[
                EntityName,
                Union[
                    Features,
                    Tuple[Features, Sequence[EntityID]],
                    None,
                ],
            ]
        ] = None,
        features: Optional[Mapping[EntityName, Features]] = None,
        ids: Optional[Mapping[EntityName, Sequence[EntityID]]] = None,
        global_features: Union[npt.NDArray[np.float32], Sequence[float], None] = None,
        actions: Optional[Mapping[ActionName, ActionMask]] = None,
        metrics: Optional[Dict[str, float]] = None,
    ) -> None:
        self.global_features = global_features if global_features is not None else []
        self.actions = actions or {}
        self.done = done
        self.reward = reward
        if features is not None:
            assert entities is None, "Cannot specify both features and entities"
            self.features = features
            self.ids = ids or {}
        else:
            self.features = {
                etype: entity[0] if isinstance(entity, tuple) else entity
                for etype, entity in (entities or {}).items()
                if entity is not None
            }
            self.ids = {
                etype: entity[1]
                for etype, entity in (entities or {}).items()
                if entity is not None and isinstance(entity, tuple)
            }
        self.metrics = metrics or {}
        self.visible = visible or {}
        self._id_to_index: Optional[Dict[EntityID, int]] = None
        self._index_to_id: Optional[List[EntityID]] = None

    @classmethod
    def empty(cls) -> "Observation":
        return Observation(
            actions={},
            done=False,
            reward=0.0,
        )

    def _actor_indices(
        self, atype: ActionName, obs_space: ObsSpace
    ) -> npt.NDArray[np.int64]:
        action = self.actions[atype]
        if isinstance(action, GlobalCategoricalActionMask):
            return np.array(
                [sum(len(v) for v in self.features.values())], dtype=np.int64
            )
        elif action.actor_ids is not None:
            id_to_index = self.id_to_index(obs_space)
            return np.array(
                [id_to_index[id] for id in action.actor_ids], dtype=np.int64
            )
        elif action.actor_types is not None:
            ids: List[int] = []
            id_to_index = self.id_to_index(obs_space)
            for etype in action.actor_types:
                ids.extend(id_to_index[id] for id in self.ids[etype])
            return np.array(ids, dtype=np.int64)
        else:
            return np.arange(
                sum(len(self.ids[etype]) for etype in obs_space.entities),
                dtype=np.int64,
            )

    def _actee_indices(
        self, atype: ActionName, obs_space: ObsSpace
    ) -> npt.NDArray[np.int64]:
        action = self.actions[atype]
        assert isinstance(action, SelectEntityActionMask)
        if action.actee_ids is not None:
            id_to_index = self.id_to_index(obs_space)
            return np.array(
                [id_to_index[id] for id in action.actee_ids], dtype=np.int64
            )
        elif action.actee_types is not None:
            ids: List[int] = []
            id_to_index = self.id_to_index(obs_space)
            for etype in action.actee_types:
                ids.extend(id_to_index[id] for id in self.ids[etype])
            return np.array(ids, dtype=np.int64)
        else:
            return np.arange(
                sum(len(self.ids[etype]) for etype in obs_space.entities),
                dtype=np.int64,
            )

    def id_to_index(self, obs_space: ObsSpace) -> Dict[EntityID, int]:
        offset = 0
        if self._id_to_index is None:
            self._id_to_index = {}
            for etype in obs_space.entities.keys():
                ids = self.ids.get(etype)
                if ids is None:
                    continue
                for i, id in enumerate(ids):
                    self._id_to_index[id] = i + offset
                offset += len(ids)
        return self._id_to_index

    def index_to_id(self, obs_space: ObsSpace) -> List[EntityID]:
        if self._index_to_id is None:
            self._index_to_id = []
            for etype in obs_space.entities.keys():
                ids = self.ids.get(etype)
                if ids is None:
                    ids = [None] * self.num_entities(etype)
                self._index_to_id.extend(ids)
        return self._index_to_id

    def num_entities(self, entity: EntityName) -> int:
        feats = self.features[entity]
        if isinstance(feats, np.ndarray):
            return feats.shape[0]
        else:
            return len(feats)


class Environment(ABC):
    """
    Abstract base class for all environments.
    """

    @abstractmethod
    def obs_space(self) -> ObsSpace:
        """
        Defines the shape of observations returned by the environment.
        """
        raise NotImplementedError

    @abstractmethod
    def action_space(self) -> Dict[str, ActionSpace]:
        """
        Defines the types of actions that can be taken in the environment.
        """
        raise NotImplementedError

    @abstractmethod
    def reset(self) -> Observation:
        """
        Resets the environment and returns the initial observation.
        """
        raise NotImplementedError

    @abstractmethod
    def act(self, actions: Mapping[ActionName, Action]) -> Observation:
        """
        Performs the given action and returns the resulting observation.

        :param actions: Maps the name of each action type to the action to perform.
        """
        raise NotImplementedError

    def reset_filter(self, obs_filter: ObsSpace) -> Observation:
        """
        Resets the environment and returns the initial observation.
        Any entities or features that are not present in the filter are removed from the observation.
        """
        return self._filter_obs(self.reset(), obs_filter)

    def render(self, **kwargs: Any) -> npt.NDArray[np.uint8]:
        """
        Renders the environment.

        :param kwargs: a dictionary of arguments to send to the rendering process
        """
        raise NotImplementedError

    def act_filter(
        self, actions: Mapping[ActionName, Action], obs_filter: ObsSpace
    ) -> Observation:
        """
        Performs the given action and returns the resulting observation.
        Any entities or features that are not present in the filter are removed from the observation.
        """
        return self._filter_obs(self.act(actions), obs_filter)

    def close(self) -> None:
        """Closes the environment."""

    def _filter_obs(self, obs: Observation, obs_filter: ObsSpace) -> Observation:
        selectors = self._compile_feature_filter(obs_filter)
        features: Dict[
            EntityName, Union[npt.NDArray[np.float32], Sequence[Sequence[float]]]
        ] = {}
        for etype, feats in obs.features.items():
            selector = selectors[etype]
            if isinstance(feats, np.ndarray):
                features[etype] = feats[:, selector].reshape(
                    feats.shape[0], len(selector)
                )
            else:
                features[etype] = [[entity[i] for i in selector] for entity in feats]
        return Observation(
            global_features=obs.global_features,
            features=features,
            ids=obs.ids,
            actions=obs.actions,
            done=obs.done,
            reward=obs.reward,
            metrics=obs.metrics,
            visible=obs.visible,
        )

    def _compile_feature_filter(self, obs_space: ObsSpace) -> Dict[str, np.ndarray]:
        obs_space = self.obs_space()
        feature_selection = {}
        for entity_name, entity in obs_space.entities.items():
            feature_selection[entity_name] = np.array(
                [entity.features.index(f) for f in entity.features], dtype=np.int32
            )
        feature_selection["__global__"] = np.array(
            [obs_space.global_features.index(f) for f in obs_space.global_features],
            dtype=np.int32,
        )
        return feature_selection



import random
from typing import Dict, Mapping, Tuple

from entity_gym.env import *
from entity_gym.runner import CliRunner


class TreasureHunt(Environment):
    def reset(self) -> Observation:
        self.x_pos = 0
        self.y_pos = 0
        self.game_over = False
        self.traps = []
        self.treasure = []
        for _ in range(5):
            self.traps.append(self._random_empty_pos())
        for _ in range(5):
            self.treasure.append(self._random_empty_pos())
        return self._observe()

    def obs_space(self) -> ObsSpace:
        return ObsSpace(
            global_features=["x_pos", "y_pos"],
            entities={
                "Trap": Entity(features=["x_pos", "y_pos"]),
                "Treasure": Entity(features=["x_pos", "y_pos"]),
            },
        )

    def action_space(self) -> Dict[str, ActionSpace]:
        # The `GlobalCategoricalActionSpace` allows the agent to choose from set of discrete actions.
        return {
            "move": GlobalCategoricalActionSpace(
                index_to_label=["up", "down", "left", "right"]
            )
        }

    def _random_empty_pos(self) -> Tuple[int, int]:
        # Generate a random position on the grid that is not occupied by a trap, treasure, or player.
        while True:
            x = random.randint(-5, 5)
            y = random.randint(-5, 5)
            if (x, y) not in (self.traps + self.treasure + [(self.x_pos, self.y_pos)]):
                return x, y

    def act(self, actions: Mapping[ActionName, Action]) -> Observation:
        action = actions["move"]
        assert isinstance(action, GlobalCategoricalAction)
        if action.label == "up" and self.y_pos < 10:
            self.y_pos += 1
        elif action.label == "down" and self.y_pos > -10:
            self.y_pos -= 1
        elif action.label == "left" and self.x_pos > -10:
            self.x_pos -= 1
        elif action.label == "right" and self.x_pos < 10:
            self.x_pos += 1

        reward = 0.0
        if (self.x_pos, self.y_pos) in self.treasure:
            reward = 1.0
            self.treasure.remove((self.x_pos, self.y_pos))
        if (self.x_pos, self.y_pos) in self.traps or len(self.treasure) == 0:
            self.game_over = True

        return self._observe(reward)

    def _observe(self, reward: float = 0.0) -> Observation:
        return Observation(
            global_features=[self.x_pos, self.y_pos],
            features={
                "Trap": self.traps,
                "Treasure": self.treasure,
            },
            done=self.game_over,
            reward=reward,
            actions={"move": GlobalCategoricalActionMask()},
        )


if __name__ == "__main__":
    env = TreasureHunt()
    CliRunner(env).run()



import random
from dataclasses import dataclass
from typing import Dict, Mapping

from entity_gym.env import Action, ActionSpace, Environment, Observation, ObsSpace
from entity_gym.env.environment import (
    Entity,
    GlobalCategoricalAction,
    GlobalCategoricalActionMask,
    GlobalCategoricalActionSpace,
)


@dataclass
class Input:
    is_set: float


class Xor(Environment):
    """
    There are three entities types, each with one instance on each timstep.
    The Bit1 and Bit2 entities are randomly set to 0 or 1.
    The Output entity has one action that should be set to the output of the XOR between the two bits.
    """

    def obs_space(self) -> ObsSpace:
        return ObsSpace(
            global_features=["negate"],
            entities={"Input": Entity(["is_set"])},
        )

    def action_space(self) -> Dict[str, ActionSpace]:
        return {"output": GlobalCategoricalActionSpace(["0", "1"])}

    def reset_filter(self, obs_space: ObsSpace) -> Observation:
        self.bit1 = random.choice([0.0, 1.0])
        self.bit2 = random.choice([0.0, 1.0])
        self.negate = random.choice([0.0, 1.0])
        return self.observe(obs_space)

    def reset(self) -> Observation:
        return self.reset_filter(self.obs_space())

    def act_filter(
        self, action: Mapping[str, Action], obs_filter: ObsSpace
    ) -> Observation:
        reward = 0.0
        a = action["output"]
        assert isinstance(a, GlobalCategoricalAction)
        if a.index == self.negate and self.bit1 == self.bit2:
            reward = 1.0
        elif a.index == 1.0 - self.negate and self.bit1 != self.bit2:
            reward = 1.0

        return self.observe(obs_filter, done=True, reward=reward)

    def act(self, actions: Mapping[str, Action]) -> Observation:
        return self.act_filter(
            actions,
            self.obs_space(),
        )

    def observe(
        self, obs_filter: ObsSpace, done: bool = False, reward: float = 0.0
    ) -> Observation:
        return Observation(
            reward=reward,
            done=done,
            features={"Input": [[self.bit1], [self.bit2]]},
            global_features=[self.negate],
            actions={
                "output": GlobalCategoricalActionMask(),
            },
        )



from dataclasses import dataclass
from typing import Any, Dict, List, Mapping, Optional

import msgpack_numpy
import numpy as np
from ragged_buffer import RaggedBufferF32, RaggedBufferI64

from entity_gym.env import ActionSpace, ObsSpace, VecEnv, VecObs
from entity_gym.env.environment import ActionName
from entity_gym.serialization.msgpack_ragged import (
    ragged_buffer_decode,
    ragged_buffer_encode,
)


@dataclass
class Sample:
    obs: VecObs
    step: List[int]
    episode: List[int]
    actions: Mapping[str, RaggedBufferI64]
    probs: Dict[str, RaggedBufferF32]
    logits: Optional[Dict[str, RaggedBufferF32]]

    def serialize(self) -> bytes:
        return msgpack_numpy.dumps(  # type: ignore
            {
                "obs": self.obs,
                "step": self.step,
                "episode": self.episode,
                "actions": self.actions,
                "probs": self.probs,
                "logits": self.logits,
            },
            default=ragged_buffer_encode,
        )

    @classmethod
    def deserialize(cls, data: bytes) -> "Sample":
        return Sample(
            **msgpack_numpy.loads(
                data, object_hook=ragged_buffer_decode, strict_map_key=False
            )
        )


class SampleRecorder:
    """
    Writes samples to disk.
    """

    def __init__(
        self,
        path: str,
        act_space: Dict[str, ActionSpace],
        obs_space: ObsSpace,
        subsample: int,
    ) -> None:
        self.path = path
        self.file = open(path, "wb")

        # Version 0
        self.file.write(np.uint64(1).tobytes())

        bytes = msgpack_numpy.dumps(
            {
                "act_space": act_space,
                "obs_space": obs_space,
                "subsample": subsample,
            },
            default=ragged_buffer_encode,
        )
        self.file.write(np.uint64(len(bytes)).tobytes())
        self.file.write(bytes)

    def record(
        self,
        sample: Sample,
    ) -> None:
        bytes = sample.serialize()
        # Write 8 bytes unsigned int for the size of the serialized sample
        self.file.write(np.uint64(len(bytes)).tobytes())
        self.file.write(bytes)

    def close(self) -> None:
        self.file.close()


class SampleRecordingVecEnv(VecEnv):
    def __init__(
        self,
        inner: VecEnv,
        out_path: str,
        subsample: int = 1,
    ) -> None:
        self.inner = inner
        self.out_path = out_path
        self.subsample = subsample
        self.sample_recorder = SampleRecorder(
            out_path,
            inner.action_space(),
            inner.obs_space(),
            subsample,
        )
        self.last_obs: Optional[VecObs] = None
        self.episodes = list(range(len(inner)))
        self.curr_step = [0] * len(inner)
        self.next_episode = len(inner)
        self.rng = np.random.default_rng(0)

    def reset(self, obs_config: ObsSpace) -> VecObs:
        self.curr_step = [0] * len(self)
        self.last_obs = self.record_obs(self.inner.reset(obs_config))
        return self.last_obs

    def record_obs(self, obs: VecObs) -> VecObs:
        for i, done in enumerate(obs.done):
            if done:
                self.episodes[i] = self.next_episode
                self.next_episode += 1
                self.curr_step[i] = 0
            else:
                self.curr_step[i] += 1
        self.last_obs = obs
        return obs

    def act(
        self,
        actions: Mapping[str, RaggedBufferI64],
        obs_filter: ObsSpace,
        probs: Optional[Dict[str, RaggedBufferF32]] = None,
        logits: Optional[Dict[str, RaggedBufferF32]] = None,
    ) -> VecObs:
        if probs is None:
            probs = {}
        # with tracer.span("record_samples"):
        assert self.last_obs is not None
        if self.subsample > 1:
            select = self.rng.integers(0, self.subsample, size=len(self.episodes)) == 0
            indices = np.arange(len(self.episodes))[select]
            if len(indices) > 0:
                last_obs = VecObs(
                    features={k: v[indices] for k, v in self.last_obs.features.items()},
                    action_masks={
                        k: v[indices] for k, v in self.last_obs.action_masks.items()
                    },
                    reward=self.last_obs.reward[select],
                    done=self.last_obs.done[select],
                    metrics=self.last_obs.metrics,
                    visible={k: v[indices] for k, v in self.last_obs.visible.items()},
                )
                self.sample_recorder.record(
                    Sample(
                        obs=last_obs,
                        step=[step for step, s in zip(self.curr_step, select) if s],
                        episode=[e for e, s in zip(self.episodes, select) if s],
                        actions={k: v[indices] for k, v in actions.items()},
                        probs={k: v[indices] for k, v in probs.items()}
                        if probs is not None
                        else None,
                        logits=(
                            {k: v[indices] for k, v in logits.items()}
                            if logits is not None
                            else None
                        ),
                    )
                )
        else:
            self.sample_recorder.record(
                Sample(
                    self.last_obs,
                    step=list(self.curr_step),
                    episode=list(self.episodes),
                    actions=actions,
                    probs=probs,
                    logits=logits,
                )
            )
        return self.record_obs(self.inner.act(actions, obs_filter))

    def render(self, **kwargs: Any) -> np.ndarray:
        return self.inner.render(**kwargs)

    def action_space(self) -> Dict[ActionName, ActionSpace]:
        return self.inner.action_space()

    def obs_space(self) -> ObsSpace:
        return self.inner.obs_space()

    def __len__(self) -> int:
        return len(self.inner)

    def close(self) -> None:
        self.sample_recorder.close()
        print("Recorded samples to: ", self.sample_recorder.path)
        self.inner.close()



import random
from dataclasses import dataclass
from typing import Dict, Mapping

from entity_gym.dataclass_utils import extract_features, obs_space_from_dataclasses
from entity_gym.env import (
    Action,
    ActionSpace,
    CategoricalAction,
    CategoricalActionMask,
    CategoricalActionSpace,
    Environment,
    Observation,
    ObsSpace,
)


@dataclass
class Lava:
    x: float
    y: float


@dataclass
class HighGround:
    x: float
    y: float


@dataclass
class Player:
    x: float
    y: float


class FloorIsLava(Environment):
    """
    The player is surrounded by 8 tiles, 7 of which are lava and 1 of which is high ground.
    The player must move to one of the tiles.
    The player receives a reward of 1 if they move to the high ground, and 0 otherwise.
    """

    def obs_space(self) -> ObsSpace:
        return obs_space_from_dataclasses(Lava, HighGround, Player)

    def action_space(self) -> Dict[str, ActionSpace]:
        return {
            "move": CategoricalActionSpace(["n", "ne", "e", "se", "s", "sw", "w", "nw"])
        }

    def reset_filter(self, obs_space: ObsSpace) -> Observation:
        width = 1000
        x = random.randint(-width, width)
        y = random.randint(-width, width)
        self.player = Player(x, y)
        self.lava = random.sample(
            [
                Lava(x + i, y + j)
                for i in range(-1, 2)
                for j in range(-1, 2)
                if not (i == 0 and j == 0)
            ],
            random.randint(1, 8),
        )
        safe = random.randint(0, len(self.lava) - 1)
        self.high_ground = HighGround(self.lava[safe].x, self.lava[safe].y)
        self.lava.pop(safe)
        obs = self.observe(obs_space)
        return obs

    def reset(self) -> Observation:
        return self.reset_filter(self.obs_space())

    def act_filter(
        self, action: Mapping[str, Action], obs_filter: ObsSpace
    ) -> Observation:
        for action_name, a in action.items():
            assert isinstance(a, CategoricalAction) and action_name == "move"
            dx, dy = [
                (0, 1),
                (1, 1),
                (1, 0),
                (1, -1),
                (0, -1),
                (-1, -1),
                (-1, 0),
                (-1, 1),
            ][a.indices[0]]
            self.player.x += dx
            self.player.y += dy
        obs = self.observe(obs_filter, done=True)
        return obs

    def act(self, actions: Mapping[str, Action]) -> Observation:
        return self.act_filter(
            actions,
            self.obs_space(),
        )

    def observe(self, obs_filter: ObsSpace, done: bool = False) -> Observation:
        if (
            done
            and self.player.x == self.high_ground.x
            and self.player.y == self.high_ground.y
        ):
            reward = 1.0
        else:
            reward = 0.0
        return Observation(
            features=extract_features(
                {
                    "Player": [self.player],
                    "Lava": self.lava,
                    "HighGround": [self.high_ground],
                },
                obs_filter,
            ),
            actions={
                "move": CategoricalActionMask(actor_types=["Player"]),
            },
            ids={"Player": [0]},
            reward=reward,
            done=done,
        )



import random
from dataclasses import dataclass
from typing import Dict, Mapping

import numpy as np

from entity_gym.env import (
    Action,
    ActionSpace,
    CategoricalAction,
    CategoricalActionMask,
    CategoricalActionSpace,
    Entity,
    Environment,
    Observation,
    ObsSpace,
)


@dataclass
class MoveToOrigin(Environment):
    """
    Task with a single Spaceship that is rewarded for moving as close to the origin as possible.
    The Spaceship has two actions for accelerating the Spaceship in the x and y directions.
    """

    x_pos: float = 0.0
    y_pos: float = 0.0
    x_velocity: float = 0.0
    y_velocity: float = 0.0
    last_x_pos = 0.0
    last_y_pos = 0.0
    step: int = 0

    def obs_space(cls) -> ObsSpace:
        return ObsSpace(
            entities={
                "Spaceship": Entity(
                    ["x_pos", "y_pos", "x_velocity", "y_velocity", "step"]
                ),
            }
        )

    def action_space(cls) -> Dict[str, ActionSpace]:
        return {
            "horizontal_thruster": CategoricalActionSpace(
                [
                    "100% right",
                    "10% right",
                    "hold",
                    "10% left",
                    "100% left",
                ],
            ),
            "vertical_thruster": CategoricalActionSpace(
                ["100% up", "10% up", "hold", "10% down", "100% down"],
            ),
        }

    def reset(self) -> Observation:
        angle = random.uniform(0, 2 * np.pi)
        self.x_pos = np.cos(angle)
        self.y_pos = np.sin(angle)
        self.last_x_pos = self.x_pos
        self.last_y_pos = self.y_pos
        self.x_velocity = 0
        self.y_velocity = 0
        self.step = 0
        return self.observe()

    def act(self, actions: Mapping[str, Action]) -> Observation:
        self.step += 1

        for action_name, a in actions.items():
            assert isinstance(a, CategoricalAction), f"{a} is not a CategoricalAction"
            if action_name == "horizontal_thruster":
                for label in a.labels:
                    if label == "100% right":
                        self.x_velocity += 0.01
                    elif label == "10% right":
                        self.x_velocity += 0.001
                    elif label == "hold":
                        pass
                    elif label == "10% left":
                        self.x_velocity -= 0.001
                    elif label == "100% left":
                        self.x_velocity -= 0.01
                    else:
                        raise ValueError(f"Invalid choice id {label}")
            elif action_name == "vertical_thruster":
                for label in a.labels:
                    if label == "100% up":
                        self.y_velocity += 0.01
                    elif label == "10% up":
                        self.y_velocity += 0.001
                    elif label == "hold":
                        pass
                    elif label == "10% down":
                        self.y_velocity -= 0.001
                    elif label == "100% down":
                        self.y_velocity -= 0.01
                    else:
                        raise ValueError(f"Invalid choice id {label}")
            else:
                raise ValueError(f"Unknown action type {action_name}")

        self.last_x_pos = self.x_pos
        self.last_y_pos = self.y_pos

        self.x_pos += self.x_velocity
        self.y_pos += self.y_velocity

        done = self.step >= 32
        return self.observe(done)

    def observe(self, done: bool = False) -> Observation:
        return Observation(
            ids={
                "Spaceship": [0],
            },
            features={
                "Spaceship": np.array(
                    [
                        [
                            self.x_pos,
                            self.y_pos,
                            self.x_velocity,
                            self.y_velocity,
                            self.step,
                        ]
                    ],
                    dtype=np.float32,
                ),
            },
            actions={
                "horizontal_thruster": CategoricalActionMask(),
                "vertical_thruster": CategoricalActionMask(),
            },
            reward=(self.last_x_pos**2 + self.last_y_pos**2) ** 0.5
            - (self.x_pos**2 + self.y_pos**2) ** 0.5,
            done=done,
        )



import copy
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, List, Mapping, Optional, Union, overload

import numpy as np
import numpy.typing as npt
from ragged_buffer import RaggedBufferBool, RaggedBufferF32, RaggedBufferI64

from entity_gym.env.environment import (
    ActionName,
    ActionSpace,
    CategoricalActionSpace,
    EntityName,
    GlobalCategoricalActionSpace,
    Observation,
    ObsSpace,
    SelectEntityActionSpace,
)


@dataclass
class VecSelectEntityActionMask:
    actors: RaggedBufferI64
    actees: RaggedBufferI64

    @overload
    def __getitem__(self, i: int) -> RaggedBufferI64:
        ...

    @overload
    def __getitem__(self, i: npt.NDArray[np.int64]) -> "VecSelectEntityActionMask":
        ...

    def __getitem__(
        self, i: Union[int, npt.NDArray[np.int64]]
    ) -> Union["VecSelectEntityActionMask", RaggedBufferI64]:
        if isinstance(i, int):
            return self.actors[i]
        else:
            return VecSelectEntityActionMask(self.actors[i], self.actees[i])

    def extend(self, other: Any) -> None:
        assert isinstance(
            other, VecSelectEntityActionMask
        ), f"Expected VecSelectEntityActionMask, got {type(other)}"
        self.actors.extend(other.actors)
        self.actees.extend(other.actees)

    def clear(self) -> None:
        self.actors.clear()
        self.actees.clear()


@dataclass
class VecCategoricalActionMask:
    actors: RaggedBufferI64
    mask: Optional[RaggedBufferBool]

    def __getitem__(
        self, i: Union[int, npt.NDArray[np.int64]]
    ) -> "VecCategoricalActionMask":
        if self.mask is not None and self.mask.size0() > 0:
            return VecCategoricalActionMask(self.actors[i], self.mask[i])
        else:
            return VecCategoricalActionMask(self.actors[i], None)

    def extend(self, other: Any) -> None:
        assert isinstance(
            other, VecCategoricalActionMask
        ), f"Expected CategoricalActionMaskBatch, got {type(other)}"
        if self.mask is not None and other.mask is not None:
            self.mask.extend(other.mask)
        elif self.mask is None and other.mask is None:
            pass
        elif self.mask is not None:
            self.mask.extend(
                RaggedBufferBool.from_flattened(
                    flattened=np.ones(
                        shape=(other.actors.items(), self.mask.size2()),
                        dtype=np.bool_,
                    ),
                    lengths=other.actors.size1(),
                )
            )
        elif other.mask is not None:
            self.mask = RaggedBufferBool.from_flattened(
                flattened=np.ones(
                    shape=(self.actors.items(), other.mask.size2()),
                    dtype=np.bool_,
                ),
                lengths=self.actors.size1(),
            )
            self.mask.extend(other.mask)
        else:
            raise Exception("Impossible!")
        self.actors.extend(other.actors)

    def clear(self) -> None:
        self.actors.clear()
        if self.mask is not None:
            self.mask.clear()


VecActionMask = Union[VecCategoricalActionMask, VecSelectEntityActionMask]


@dataclass
class Metric:
    count: int = 0
    sum: float = 0.0
    min: float = float("inf")
    max: float = float("-inf")

    def push(self, value: float) -> None:
        self.count += 1
        self.sum += value
        self.min = min(self.min, value)
        self.max = max(self.max, value)

    def __iadd__(self, m: "Metric") -> "Metric":
        self.count += m.count
        self.sum += m.sum
        self.min = min(self.min, m.min)
        self.max = max(self.max, m.max)
        return self

    def __add__(self, m: "Metric") -> "Metric":
        return Metric(
            count=self.count + m.count,
            sum=self.sum + m.sum,
            min=min(self.min, m.min),
            max=max(self.max, m.max),
        )

    @property
    def mean(self) -> float:
        if self.count == 0:
            return 0.0
        else:
            return self.sum / self.count


@dataclass
class VecObs:
    """
    A batch of observations from a vectorized environment.
    """

    features: Dict[EntityName, RaggedBufferF32]
    # Optional mask to hide specific entities from the policy but not the value function
    visible: Dict[EntityName, RaggedBufferBool]
    action_masks: Dict[ActionName, VecActionMask]
    reward: npt.NDArray[np.float32]
    done: npt.NDArray[np.bool_]
    metrics: Dict[str, Metric]

    def extend(self, b: "VecObs") -> None:
        num_envs = len(self.reward)
        # Extend visible (must happen before features in case of backfill)
        for etype in self.features.keys():
            if etype in b.visible:
                if etype not in self.visible:
                    self.visible[etype] = RaggedBufferBool.from_flattened(
                        flattened=np.ones(
                            shape=(self.features[etype].items(), 1), dtype=np.bool_
                        ),
                        lengths=self.features[etype].size1(),
                    )
                self.visible[etype].extend(b.visible[etype])
            elif etype in self.visible:
                self.visible[etype].extend(
                    RaggedBufferBool.from_flattened(
                        flattened=np.ones(
                            shape=(b.features[etype].items(), 1), dtype=np.bool_
                        ),
                        lengths=b.features[etype].size1(),
                    )
                )
        for etype, feats in b.features.items():
            if etype not in self.features:
                self.features[etype] = empty_ragged_f32(
                    feats=feats.size2(), sequences=num_envs
                )
            self.features[etype].extend(feats)
        for etype, feats in self.features.items():
            if etype not in b.features:
                feats.extend(empty_ragged_f32(feats.size2(), len(b.reward)))
        for atype, amask in b.action_masks.items():
            if atype not in self.action_masks:
                raise NotImplementedError()
            else:
                self.action_masks[atype].extend(amask)
        self.reward = np.concatenate((self.reward, b.reward))
        self.done = np.concatenate((self.done, b.done))
        num_envs = len(self.reward)
        for name, stats in b.metrics.items():
            if name in self.metrics:
                self.metrics[name].count += stats.count
                self.metrics[name].sum += stats.sum
                self.metrics[name].min = min(self.metrics[name].min, stats.min)
                self.metrics[name].max = max(self.metrics[name].max, stats.max)
            else:
                self.metrics[name] = copy.copy(stats)


class VecEnv(ABC):
    """
    Interface for vectorized environments. The main goal of VecEnv is to allow
    for maximally efficient environment implementations.
    """

    @abstractmethod
    def obs_space(self) -> ObsSpace:
        """
        Returns a dictionary mapping the name of observable entities to their type.
        """
        raise NotImplementedError

    @abstractmethod
    def action_space(self) -> Dict[ActionName, ActionSpace]:
        """
        Returns a dictionary mapping the name of actions to their action space.
        """
        raise NotImplementedError

    @abstractmethod
    def reset(self, obs_config: ObsSpace) -> VecObs:
        """
        Resets all environments and returns the initial observations.
        """
        raise NotImplementedError

    @abstractmethod
    def act(
        self, actions: Mapping[ActionName, RaggedBufferI64], obs_filter: ObsSpace
    ) -> VecObs:
        """
        Performs the given actions on the underlying environments and returns the resulting observations.
        Any environment that reaches the end of its episode is reset and returns the initial observation of the next episode.
        """
        raise NotImplementedError

    @abstractmethod
    def render(self, **kwargs: Any) -> npt.NDArray[np.uint8]:
        raise NotImplementedError

    @abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError

    def close(self) -> None:
        pass


def batch_obs(
    obs: List[Observation], obs_space: ObsSpace, action_space: Dict[str, ActionSpace]
) -> VecObs:
    """
    Converts a list of observations into a batch of observations.
    """
    features: Dict[EntityName, RaggedBufferF32] = {}
    visible: Dict[EntityName, RaggedBufferBool] = {}
    action_masks: Dict[ActionName, VecActionMask] = {}
    reward = []
    done = []
    metrics = {}

    # Initialize the entire batch with all entities and actions
    for entity_name, entity in obs_space.entities.items():
        nfeat = len(entity.features)
        features[entity_name] = RaggedBufferF32(nfeat)
    global_entity = len(obs_space.global_features) > 0
    for action_name, space in action_space.items():
        if isinstance(space, CategoricalActionSpace):
            action_masks[action_name] = VecCategoricalActionMask(
                RaggedBufferI64(1),
                None,
            )
        elif isinstance(space, GlobalCategoricalActionSpace):
            action_masks[action_name] = VecCategoricalActionMask(
                RaggedBufferI64(1),
                None,
            )
            global_entity = True
        elif isinstance(space, SelectEntityActionSpace):
            action_masks[action_name] = VecSelectEntityActionMask(
                RaggedBufferI64(1), RaggedBufferI64(1)
            )
        else:
            raise NotImplementedError(f"Action space {space} not supported")
    if global_entity:
        nfeat = len(obs_space.global_features)
        features["__global__"] = RaggedBufferF32(nfeat)

    for i, o in enumerate(obs):
        # Merge entity features
        for entity_type, entity in obs_space.entities.items():
            if entity_type not in features:
                features[entity_type] = RaggedBufferF32.from_flattened(
                    np.zeros((0, len(entity.features)), dtype=np.float32),
                    lengths=np.zeros(i, dtype=np.int64),
                )
            if entity_type in o.features:
                ofeats = o.features[entity_type]
                if not isinstance(ofeats, np.ndarray):
                    ofeats = np.array(ofeats, dtype=np.float32).reshape(
                        len(ofeats), len(obs_space.entities[entity_type].features)
                    )
                features[entity_type].push(ofeats)
            else:
                features[entity_type].push(
                    np.zeros((0, len(entity.features)), dtype=np.float32)
                )
        if global_entity:
            gfeats = o.global_features
            if not isinstance(gfeats, np.ndarray):
                gfeats = np.array(gfeats, dtype=np.float32)
            features["__global__"].push(
                gfeats.reshape(1, len(obs_space.global_features))
            )

        # Merge visibilities
        for etype, vis in o.visible.items():
            if etype not in visible:
                lengths = []
                for j in range(i):
                    if etype in obs[j].features:
                        lengths.append(len(obs[j].features[etype]))
                    else:
                        lengths.append(0)
                visible[etype] = RaggedBufferBool.from_flattened(
                    np.ones((sum(lengths), 1), dtype=np.bool_),
                    lengths=np.array(lengths, dtype=np.int64),
                )
            if not isinstance(vis, np.ndarray):
                vis = np.array(vis, dtype=np.bool_)
            visible[etype].push(vis.reshape(-1, 1))

        # Merge action masks
        for atype, space in action_space.items():
            if atype not in o.actions:
                if atype in action_masks:
                    if isinstance(space, CategoricalActionSpace):
                        vec_action = action_masks[atype]
                        assert isinstance(vec_action, VecCategoricalActionMask)
                        vec_action.actors.push(np.zeros((0, 1), dtype=np.int64))
                        if vec_action.mask is not None:
                            vec_action.mask.push(
                                np.zeros((0, len(space.index_to_label)), dtype=np.bool_)
                            )
                    elif isinstance(space, SelectEntityActionSpace):
                        vec_action = action_masks[atype]
                        assert isinstance(vec_action, VecSelectEntityActionMask)
                        vec_action.actors.push(np.zeros((0, 1), dtype=np.int64))
                        vec_action.actees.push(np.zeros((0, 1), dtype=np.int64))
                    else:
                        raise ValueError(
                            f"Unsupported action space type: {type(space)}"
                        )
                continue
            action = o.actions[atype]
            if atype not in action_masks:
                if isinstance(space, CategoricalActionSpace) or isinstance(
                    space, GlobalCategoricalActionSpace
                ):
                    action_masks[atype] = VecCategoricalActionMask(
                        empty_ragged_i64(1, i), None
                    )
                elif isinstance(space, SelectEntityActionSpace):
                    action_masks[atype] = VecSelectEntityActionMask(
                        empty_ragged_i64(1, i), empty_ragged_i64(1, i)
                    )
                else:
                    raise ValueError(f"Unknown action space type: {space}")
            if isinstance(space, CategoricalActionSpace) or isinstance(
                space, GlobalCategoricalActionSpace
            ):
                vec_action = action_masks[atype]
                assert isinstance(vec_action, VecCategoricalActionMask)
                actor_indices = o._actor_indices(atype, obs_space)
                vec_action.actors.push(actor_indices.reshape(-1, 1))
                if action.mask is not None:
                    if vec_action.mask is None:
                        vec_action.mask = RaggedBufferBool.from_flattened(
                            np.ones((0, len(space.index_to_label)), dtype=np.bool_),
                            np.zeros(i, dtype=np.int64),
                        )
                    amask = action.mask
                    if not isinstance(amask, np.ndarray):
                        amask = np.array(amask, dtype=np.bool_)
                    vec_action.mask.push(amask)
                elif vec_action.mask is not None:
                    vec_action.mask.push(
                        np.ones(
                            (len(actor_indices), len(space.index_to_label)),
                            dtype=np.bool_,
                        )
                    )
            elif isinstance(space, SelectEntityActionSpace):
                vec_action = action_masks[atype]
                assert isinstance(vec_action, VecSelectEntityActionMask)
                actors = o._actor_indices(atype, obs_space).reshape(-1, 1)
                vec_action.actors.push(actors)
                if len(actors) > 0:
                    vec_action.actees.push(
                        o._actee_indices(atype, obs_space).reshape(-1, 1)
                    )
                else:
                    vec_action.actees.push(np.zeros((0, 1), dtype=np.int64))
            else:
                raise NotImplementedError()

        reward.append(o.reward)
        done.append(o.done)
        for name, value in o.metrics.items():
            if name not in metrics:
                metrics[name] = Metric()
            metrics[name].push(value)

    return VecObs(
        features,
        visible,
        action_masks,
        np.array(reward, dtype=np.float32),
        np.array(done, dtype=np.bool_),
        metrics,
    )


def empty_ragged_f32(feats: int, sequences: int) -> RaggedBufferF32:
    return RaggedBufferF32.from_flattened(
        np.zeros((0, feats), dtype=np.float32),
        lengths=np.array([0] * sequences, dtype=np.int64),
    )


def empty_ragged_i64(feats: int, sequences: int) -> RaggedBufferI64:
    return RaggedBufferI64.from_flattened(
        np.zeros((0, feats), dtype=np.int64),
        lengths=np.array([0] * sequences, dtype=np.int64),
    )



from typing import Dict

import numpy as np
from ragged_buffer import RaggedBufferBool, RaggedBufferF32, RaggedBufferI64

from entity_gym.env import (
    ActionSpace,
    CategoricalActionMask,
    CategoricalActionSpace,
    Entity,
    Observation,
    ObsSpace,
    SelectEntityActionMask,
    SelectEntityActionSpace,
    VecCategoricalActionMask,
    VecObs,
    VecSelectEntityActionMask,
)
from entity_gym.env.env_list import EnvList
from entity_gym.env.parallel_env_list import ParallelEnvList
from entity_gym.env.vec_env import batch_obs
from entity_gym.examples.cherry_pick import CherryPick
from entity_gym.examples.xor import Xor


def test_env_list() -> None:
    # 100 environments
    envs = EnvList(CherryPick, 100)
    obs_space = envs.obs_space()

    obs_reset = envs.reset(obs_space)
    assert len(obs_reset.done) == 100

    actions = {
        "Pick Cherry": RaggedBufferI64.from_array(np.zeros((100, 1, 1), np.int64))
    }
    obs_act = envs.act(actions, obs_space)

    assert len(obs_act.done) == 100


def test_parallel_env_list() -> None:
    # 100 environments split across 10 processes
    envs = ParallelEnvList(CherryPick, 100, 10)
    obs_space = envs.obs_space()

    obs_reset = envs.reset(obs_space)
    assert len(obs_reset.done) == 100

    actions = {
        "Pick Cherry": RaggedBufferI64.from_array(np.zeros((100, 1, 1), np.int64))
    }
    obs_act = envs.act(actions, obs_space)
    assert len(obs_act.done) == 100

    envs = ParallelEnvList(Xor, 100, 10)
    obs_reset = envs.reset(envs.obs_space())
    assert len(obs_reset.done) == 100
    actions_xor = {
        "output": RaggedBufferI64.from_array(np.zeros((100, 1, 1), np.int64))
    }
    obs_act = envs.act(actions_xor, envs.obs_space())
    assert len(obs_act.done) == 100


def test_batch_obs_entities() -> None:
    """
    We  have a set of observations and only a single one of those observations contains a paricular entity.
    When this entity is batched, it needs to needs to contain 0-length rows for that entity for all other observations.
    """

    obs_space = ObsSpace(
        entities={
            "entity1": Entity(["x", "y", "z"]),
            "rare": Entity(["x", "y", "z", "health", "thing"]),
        }
    )

    action_space: Dict[str, ActionSpace] = {}

    observation1 = Observation(
        features={"entity1": np.array([[10, 10, 10], [10, 10, 10]], np.float32)},
        ids={"entity1": ["entity1_0", "entity1_1"]},
        actions={},
        reward=0.0,
        done=False,
    )

    observation2 = Observation(
        features={"entity1": np.array([[10, 10, 10]], np.float32)},
        ids={"entity1": ["entity1_0"]},
        actions={},
        reward=0.0,
        done=False,
    )

    observation3 = Observation(
        features={
            "rare": np.array(
                [[10, 10, 10, 4, 2], [10, 10, 10, 4, 2], [10, 10, 10, 4, 2]], np.float32
            )
        },
        ids={"rare": ["rare1_0", "rare1_1", "rare1_2"]},
        actions={},
        reward=0.0,
        done=False,
    )

    obs_batch = batch_obs(
        [observation1, observation2, observation3], obs_space, action_space
    )

    # entity1 observations should have a ragged array with lengths [2, 1, 0]
    # rare observations should have a ragged array with lengths [0, 0, 3]
    assert np.all(obs_batch.features["entity1"].size1() == np.array([2, 1, 0]))
    assert np.all(obs_batch.features["rare"].size1() == np.array([0, 0, 3]))


def test_batch_obs_select_entity_action() -> None:
    """
    We have three actions types that are dependent on the entities that are present in each observation
    This is common in procedurally generated environments, where types of objects/entities can be generated randomly,
    or in environments where there are possible rare interactions between entities.
    """

    obs_space = ObsSpace(
        entities={
            "entity1": Entity(["x", "y", "z"]),
            "entity2": Entity(["x", "y", "z"]),
            "entity3": Entity(["x", "y", "z"]),
        }
    )

    action_space: Dict[str, ActionSpace] = {
        "high_five": SelectEntityActionSpace(),
        "mid_five": SelectEntityActionSpace(),
        "low_five": SelectEntityActionSpace(),
    }

    observation1 = Observation(
        features={
            "entity1": np.array([[10, 10, 10]], np.float32),
            "entity2": np.array([[10, 10, 10]], np.float32),
        },
        ids={"entity1": ["entity1_0"], "entity2": ["entity2_0"]},
        actions={
            # entity1 can low five entity 2 and vice versa
            "low_five": SelectEntityActionMask(
                actor_ids=["entity1_0", "entity2_0"],
                actee_ids=["entity2_0", "entity1_0"],
            )
        },
        reward=0.0,
        done=False,
    )

    observation2 = Observation(
        features={
            "entity1": np.array([[10, 10, 10]], np.float32),
            "entity3": np.array([[10, 10, 10]], np.float32),
        },
        ids={"entity1": ["entity1_0"], "entity3": ["entity3_0"]},
        actions={
            # entity3 can high five entity 1
            "high_five": SelectEntityActionMask(
                actor_ids=["entity3_0"],
                actee_ids=["entity1_0"],
            )
        },
        reward=0.0,
        done=False,
    )

    observation3 = Observation(
        features={
            "entity1": np.array([[10, 10, 10]], np.float32),
            "entity2": np.array([[10, 10, 10], [10, 10, 10]], np.float32),
            "entity3": np.array([[10, 10, 10]], np.float32),
        },
        ids={
            "entity1": ["entity1_0"],
            "entity2": ["entity2_1", "entity2_0"],
            "entity3": ["entity3_0"],
        },
        actions={
            # entity3 can high five entity 1, and entity 2_0 and entity 2_1 can mid five entity3. entity1 and entity2 can low five each other
            "high_five": SelectEntityActionMask(
                actor_ids=["entity3_0"],
                actee_ids=["entity1_0"],
            ),
            "mid_five": SelectEntityActionMask(
                actor_ids=["entity2_1", "entity2_0"],
                actee_ids=["entity3_0"],
            ),
            "low_five": SelectEntityActionMask(
                actor_ids=["entity1_0", "entity2_1", "entity2_0"],
                actee_ids=["entity2_1", "entity1_0", "entity2_0"],
            ),
        },
        reward=0.0,
        done=False,
    )

    obs_batch = batch_obs(
        [observation1, observation2, observation3], obs_space, action_space
    )

    assert isinstance(obs_batch.action_masks["high_five"], VecSelectEntityActionMask)
    assert isinstance(obs_batch.action_masks["mid_five"], VecSelectEntityActionMask)
    assert isinstance(obs_batch.action_masks["low_five"], VecSelectEntityActionMask)

    assert np.all(
        obs_batch.action_masks["high_five"].actors.size1() == np.array([0, 1, 1])
    )
    assert np.all(
        obs_batch.action_masks["high_five"].actees.size1() == np.array([0, 1, 1])
    )

    assert np.all(
        obs_batch.action_masks["mid_five"].actors.size1() == np.array([0, 0, 2])
    )
    assert np.all(
        obs_batch.action_masks["mid_five"].actees.size1() == np.array([0, 0, 1])
    )

    assert np.all(
        obs_batch.action_masks["low_five"].actors.size1() == np.array([2, 0, 3])
    )
    assert np.all(
        obs_batch.action_masks["low_five"].actees.size1() == np.array([2, 0, 3])
    )


def test_batch_obs_categorical_action() -> None:
    """
    In some cases there are categorical that may only exist for certain entity types, or may only exist under certain circumstances.
    A particular example would be an action that is only available when an entity has a particular state (a special item or similar)
    """

    obs_space = ObsSpace(
        entities={
            "entity1": Entity(["x", "y", "z"]),
            "entity2": Entity(["x", "y", "z"]),
            "entity3": Entity(["x", "y", "z"]),
        }
    )

    action_space: Dict[str, ActionSpace] = {
        "move": CategoricalActionSpace(["up", "down", "left", "right"]),
        "choose_inventory_item": CategoricalActionSpace(["axe", "sword", "pigeon"]),
    }

    observation1 = Observation(
        features={
            "entity1": np.array([[10, 10, 10]], np.float32),
            "entity2": np.array([[10, 10, 10]], np.float32),
        },
        ids={
            "entity1": ["entity1_0"],
            "entity2": ["entity2_0"],
        },
        actions={
            # both entity1 and entity2 can move all directions
            "move": CategoricalActionMask(
                actor_ids=["entity1_0", "entity2_0"],
                mask=np.array([[True, True, True, True], [True, True, True, True]]),
            ),
        },
        reward=0.0,
        done=False,
    )

    observation2 = Observation(
        features={
            "entity1": np.array([[10, 10, 10]], np.float32),
            "entity3": np.array([[10, 10, 10], [10, 10, 10]], np.float32),
        },
        ids={
            "entity1": ["entity1_0"],
            "entity3": ["entity3_0", "entity3_1"],
        },
        actions={
            # all entities can move. Entity 3_1 can also choose items
            "move": CategoricalActionMask(
                actor_ids=["entity3_0", "entity1_0", "entity3_1"],
                mask=None,
            ),
            "choose_inventory_item": CategoricalActionMask(
                ["entity3_1"], mask=np.array([[True, True, True]])
            ),
        },
        reward=0.0,
        done=False,
    )

    observation3 = Observation(
        features={
            "entity1": np.array([[10, 10, 10], [10, 10, 10]], np.float32),
            "entity2": np.array([[10, 10, 10], [10, 10, 10]], np.float32),
            "entity3": np.array([[10, 10, 10], [10, 10, 10]], np.float32),
        },
        ids={
            "entity1": ["entity1_0", "entity1_1"],
            "entity2": ["entity2_0", "entity2_1"],
            "entity3": ["entity3_1", "entity3_0"],
        },
        actions={
            # no entities can move or do anything
        },
        reward=0.0,
        done=False,
    )

    obs_batch = batch_obs(
        [observation1, observation2, observation3], obs_space, action_space
    )

    assert isinstance(obs_batch.action_masks["move"], VecCategoricalActionMask)
    assert isinstance(
        obs_batch.action_masks["choose_inventory_item"], VecCategoricalActionMask
    )

    assert obs_batch.action_masks["move"].mask is not None
    assert obs_batch.action_masks["choose_inventory_item"].mask is not None

    assert np.array_equal(
        obs_batch.action_masks["move"].actors.size1(), np.array([2, 3, 0])
    )
    assert np.array_equal(
        obs_batch.action_masks["move"].mask.size1(), np.array([2, 3, 0])
    )

    assert np.all(
        obs_batch.action_masks["choose_inventory_item"].actors.size1()
        == np.array([0, 1, 0])
    )
    assert np.all(
        obs_batch.action_masks["choose_inventory_item"].mask.size1()
        == np.array([0, 1, 0])
    )


def test_merge_obs_entities() -> None:
    """
    First batch has only entity1 and second batch only has entity2.

    both batches have 10 entries (some are 0-length)

    the output batch should have 20 of EACH entity 1 and entity2, but with zero length rows padded appropriately
    """

    obs_batch1 = VecObs(
        features={
            "entity1": RaggedBufferF32.from_flattened(
                flattened=np.array([[10, 10, 10]] * 10, np.float32),
                lengths=np.array([0, 0, 1, 1, 0, 2, 3, 3, 0, 0]),
            ),
            "entity2": RaggedBufferF32.from_flattened(
                flattened=np.array([], np.float32).reshape(0, 3),
                lengths=np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
            ),
        },
        action_masks={},
        reward=np.array([0] * 10, np.float32),
        done=np.array([False] * 10, np.bool_),
        metrics={},
        visible={},
    )

    obs_batch2 = VecObs(
        features={
            "entity1": RaggedBufferF32.from_flattened(
                flattened=np.array([], np.float32).reshape(0, 3),
                lengths=np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
            ),
            "entity2": RaggedBufferF32.from_flattened(
                flattened=np.array([[10, 10, 10]] * 10, np.float32),
                lengths=np.array([0, 0, 1, 1, 0, 2, 3, 3, 0, 0]),
            ),
        },
        action_masks={},
        reward=np.array([0] * 10, np.float32),
        done=np.array([False] * 10, np.bool_),
        metrics={},
        visible={},
    )

    VecObs.extend(obs_batch1, obs_batch2)

    assert np.all(
        obs_batch1.features["entity1"].size1()
        == [0, 0, 1, 1, 0, 2, 3, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    )
    assert np.all(
        obs_batch1.features["entity2"].size1()
        == [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 2, 3, 3, 0, 0]
    )


def test_merge_obs_actions_categorical() -> None:
    """
    There are 3 of entity 1 and 3 of entity 2.

    We have batches which contain two categorical actions "action1" and "action2"
    The first batch contains 4 action1 in 3 observations (2 in the second obsercation) and none of action2
    The second batch contains 4 action2 in 3 observations (2 in the second observation) and none of action1

    We test that when merging these two batches, the number of empty action rows is consistent in the super-batch
    action2 is padded to 0-length rows in the first batch and action1 is padded with 0-length rows in the second batch.
    Overall batch1 and batch 2 should contain 6 observations for both action1 and action2 (with 0-length rows where appropriate)
    """

    obs_batch1 = VecObs(
        features={
            "entity1": RaggedBufferF32.from_flattened(
                flattened=np.array([[10, 10, 10]] * 3, np.float32),
                lengths=np.array([1, 1, 1]),
            ),
            "entity2": RaggedBufferF32.from_flattened(
                flattened=np.array([[10, 10, 10]] * 3, np.float32),
                lengths=np.array([1, 1, 1]),
            ),
        },
        action_masks={
            "action1": VecCategoricalActionMask(
                RaggedBufferI64.from_flattened(
                    flattened=np.array([[0], [1], [2], [3]], int),
                    lengths=np.array([1, 2, 1]),
                ),
                RaggedBufferBool.from_flattened(
                    flattened=np.array(
                        [
                            [True, True, True, True],
                            [True, True, True, True],
                            [True, True, True, True],
                            [True, True, True, True],
                        ],
                        np.bool_,
                    ),
                    lengths=np.array([1, 2, 1]),
                ),
            ),
            "action2": VecCategoricalActionMask(
                RaggedBufferI64.from_flattened(
                    flattened=np.array([], int).reshape(0, 1),
                    lengths=np.array([0, 0, 0]),
                ),
                RaggedBufferBool.from_flattened(
                    flattened=np.array([], np.bool_).reshape(0, 4),
                    lengths=np.array([0, 0, 0]),
                ),
            ),
        },
        reward=np.array([0] * 10, np.float32),
        done=np.array([False] * 10, np.bool_),
        metrics={},
        visible={},
    )

    obs_batch2 = VecObs(
        features={
            "entity1": RaggedBufferF32.from_flattened(
                flattened=np.array([[10, 10, 10]] * 3, np.float32),
                lengths=np.array([1, 1, 1]),
            ),
            "entity2": RaggedBufferF32.from_flattened(
                flattened=np.array([[10, 10, 10]] * 3, np.float32),
                lengths=np.array([1, 1, 1]),
            ),
        },
        action_masks={
            "action2": VecCategoricalActionMask(
                RaggedBufferI64.from_flattened(
                    flattened=np.array([[0], [0], [1], [0]], int),
                    lengths=np.array([1, 2, 1]),
                ),
                RaggedBufferBool.from_flattened(
                    flattened=np.array(
                        [
                            [True, True, True, True],
                            [True, True, True, True],
                            [True, True, True, True],
                            [True, True, True, True],
                        ],
                        np.bool_,
                    ),
                    lengths=np.array([1, 2, 1]),
                ),
            ),
            "action1": VecCategoricalActionMask(
                RaggedBufferI64.from_flattened(
                    flattened=np.array([], int).reshape(0, 1),
                    lengths=np.array([0, 0, 0]),
                ),
                RaggedBufferBool.from_flattened(
                    flattened=np.array([], np.bool_).reshape(0, 4),
                    lengths=np.array([0, 0, 0]),
                ),
            ),
        },
        reward=np.array([0] * 10, np.float32),
        done=np.array([False] * 10, np.bool_),
        metrics={},
        visible={},
    )

    VecObs.extend(obs_batch1, obs_batch2)

    assert isinstance(obs_batch1.action_masks["action1"], VecCategoricalActionMask)
    assert isinstance(obs_batch1.action_masks["action2"], VecCategoricalActionMask)

    assert np.all(
        obs_batch1.action_masks["action1"].actors.size1() == [1, 2, 1, 0, 0, 0]
    )
    assert np.all(
        obs_batch1.action_masks["action2"].actors.size1() == [0, 0, 0, 1, 2, 1]
    )


def test_merge_obs_actions_select_entity() -> None:
    """
    There are 3 of entity 1 and 3 of entity 2.

    We have batches which contain two select entity actions "action1" and "action2"
    The first batch contains 3 observations of action1 with 1, 2 and 1 actors and 2, 1 and 2 actees respectively
    The first batch contains 3 observations of action2 with 1, 2 and 1 actors and 2, 1 and 2 actees respectively

    We test that when merging these two batches, the number of empty action rows is consistent in the super-batch
    action2 is padded to 0-length rows in the first batch and action1 is padded with 0-length rows in the second batch.
    Overall batch1 and batch 2 should contain 6 observations for both action1 and action2 (with 0-length rows where appropriate)
    """
    obs_batch1 = VecObs(
        features={
            "entity1": RaggedBufferF32.from_flattened(
                flattened=np.array([[10, 10, 10]] * 3, np.float32),
                lengths=np.array([1, 1, 1]),
            ),
            "entity2": RaggedBufferF32.from_flattened(
                flattened=np.array([[10, 10, 10]] * 3, np.float32),
                lengths=np.array([1, 1, 1]),
            ),
        },
        action_masks={
            "action1": VecSelectEntityActionMask(
                RaggedBufferI64.from_flattened(
                    flattened=np.array([[0], [0], [1], [2]], int),
                    lengths=np.array([1, 2, 1]),
                ),
                RaggedBufferI64.from_flattened(
                    flattened=np.array([[3], [4], [5], [6], [0]], int),
                    lengths=np.array([2, 1, 2]),
                ),
            ),
            "action2": VecSelectEntityActionMask(
                RaggedBufferI64.from_flattened(
                    flattened=np.array([], int).reshape(0, 1),
                    lengths=np.array([0, 0, 0]),
                ),
                RaggedBufferI64.from_flattened(
                    flattened=np.array([], int).reshape(0, 1),
                    lengths=np.array([0, 0, 0]),
                ),
            ),
        },
        reward=np.array([0] * 10, np.float32),
        done=np.array([False] * 10, np.bool_),
        metrics={},
        visible={},
    )

    obs_batch2 = VecObs(
        features={
            "entity1": RaggedBufferF32.from_flattened(
                flattened=np.array([[10, 10, 10]] * 3, np.float32),
                lengths=np.array([1, 1, 1]),
            ),
            "entity2": RaggedBufferF32.from_flattened(
                flattened=np.array([[10, 10, 10]] * 3, np.float32),
                lengths=np.array([1, 1, 1]),
            ),
        },
        action_masks={
            "action1": VecSelectEntityActionMask(
                RaggedBufferI64.from_flattened(
                    flattened=np.array([], int).reshape(0, 1),
                    lengths=np.array([0, 0, 0]),
                ),
                RaggedBufferI64.from_flattened(
                    flattened=np.array([], int).reshape(0, 1),
                    lengths=np.array([0, 0, 0]),
                ),
            ),
            "action2": VecSelectEntityActionMask(
                RaggedBufferI64.from_flattened(
                    flattened=np.array([[0], [0], [1], [2]], int),
                    lengths=np.array([1, 2, 1]),
                ),
                RaggedBufferI64.from_flattened(
                    flattened=np.array([[3], [4], [5], [6], [0]], int),
                    lengths=np.array([2, 1, 2]),
                ),
            ),
        },
        reward=np.array([0] * 10, np.float32),
        done=np.array([False] * 10, np.bool_),
        metrics={},
        visible={},
    )

    obs_batch1.extend(obs_batch2)

    assert isinstance(obs_batch1.action_masks["action1"], VecSelectEntityActionMask)
    assert isinstance(obs_batch1.action_masks["action2"], VecSelectEntityActionMask)

    assert np.all(
        obs_batch1.action_masks["action1"].actors.size1() == [1, 2, 1, 0, 0, 0]
    )
    assert np.all(
        obs_batch1.action_masks["action1"].actees.size1() == [2, 1, 2, 0, 0, 0]
    )
    assert np.all(
        obs_batch1.action_masks["action2"].actors.size1() == [0, 0, 0, 1, 2, 1]
    )
    assert np.all(
        obs_batch1.action_masks["action2"].actees.size1() == [0, 0, 0, 2, 1, 2]
    )


def test_merge_empty_masks() -> None:
    obs_batch1 = VecObs(
        features={
            "archer": RaggedBufferF32.from_flattened(
                flattened=np.array([[10, 10, 10]] * 4, np.float32),
                lengths=np.array([1, 2, 1]),
            ),
        },
        action_masks={
            "shoot": VecCategoricalActionMask(
                actors=RaggedBufferI64.from_flattened(
                    flattened=np.array([[0], [0], [1]], int),
                    lengths=np.array([1, 2, 0]),
                ),
                mask=RaggedBufferBool.from_flattened(
                    flattened=np.array(
                        [[True, True], [True, False], [False, True]], bool
                    ),
                    lengths=np.array([1, 2, 0]),
                ),
            ),
        },
        reward=np.array([0] * 3, np.float32),
        done=np.array([False] * 3, np.bool_),
        metrics={},
        visible={},
    )

    obs_batch2 = VecObs(
        features={
            "archer": RaggedBufferF32.from_flattened(
                flattened=np.array([[10, 10, 10]] * 4, np.float32),
                lengths=np.array([0, 1, 3]),
            ),
        },
        action_masks={
            "shoot": VecCategoricalActionMask(
                actors=RaggedBufferI64.from_flattened(
                    flattened=np.array([[0], [1], [2]], int),
                    lengths=np.array([0, 0, 3]),
                ),
                mask=None,
            ),
        },
        reward=np.array([0] * 3, np.float32),
        done=np.array([False] * 3, np.bool_),
        metrics={},
        visible={},
    )

    obs_batch1.extend(obs_batch2)
    obs_batch2.extend(obs_batch1)

    assert np.array_equal(
        obs_batch2.action_masks["shoot"].mask.size1(),  # type: ignore
        np.array([0, 0, 3, 1, 2, 0, 0, 0, 3]),
    ), f"{obs_batch2.action_masks['shoot'].mask.size1()}"  # type: ignore
    assert np.array_equal(
        obs_batch2.action_masks["shoot"].mask.as_array(),  # type: ignore
        np.array(
            [
                [True, True],
                [True, True],
                [True, True],
                [True, True],
                [True, False],
                [False, True],
                [True, True],
                [True, True],
                [True, True],
            ],
            bool,
        ),
    )



import multiprocessing as mp
import multiprocessing.connection as conn
from multiprocessing.connection import Connection
from typing import Any, Callable, Dict, Generator, List, Mapping, Optional

import cloudpickle
import msgpack_numpy
import numpy as np
import numpy.typing as npt
from ragged_buffer import RaggedBufferI64

from entity_gym.env.env_list import EnvList
from entity_gym.env.environment import (
    ActionName,
    ActionSpace,
    Environment,
    Observation,
    ObsSpace,
)
from entity_gym.env.vec_env import VecEnv, VecObs, batch_obs
from entity_gym.serialization.msgpack_ragged import (
    ragged_buffer_decode,
    ragged_buffer_encode,
)


class CloudpickleWrapper:
    def __init__(self, var: Any):
        self.var = var

    def __getstate__(self) -> Any:
        return cloudpickle.dumps(self.var)

    def __setstate__(self, var: Any) -> None:
        self.var = cloudpickle.loads(var)


class MsgpackConnectionWrapper:
    """
    Use msgpack instead of pickle to send and receive data from workers.
    """

    def __init__(self, conn: Connection) -> None:
        self._conn = conn

    def close(self) -> None:
        self._conn.close()

    def send(self, data: Any) -> None:
        s = msgpack_numpy.dumps(data, default=ragged_buffer_encode)
        self._conn.send_bytes(s)

    def recv(self) -> Any:
        data_bytes = self._conn.recv_bytes()
        return msgpack_numpy.loads(
            data_bytes,
            object_hook=ragged_buffer_decode,
            strict_map_key=False,
        )


def _worker(
    remote: conn.Connection,
    parent_remote: conn.Connection,
    env_list_config: CloudpickleWrapper,
) -> None:
    parent_remote.close()
    env_args = env_list_config.var
    envs = EnvList(*env_args)
    while True:
        try:
            cmd, data = remote.recv()
            if cmd == "act":
                observation = envs.act(data[0], data[1])
                remote.send(observation)
            elif cmd == "reset":
                observation = envs.reset(data)
                remote.send(observation)
            elif cmd == "render":
                rgb_pixels = envs.render(**data)
                remote.send(rgb_pixels)
            elif cmd == "close":
                envs.close()
                remote.close()
                break
            else:
                raise NotImplementedError(f"`{cmd}` is not implemented in the worker")
        except EOFError:
            break


class ParallelEnvList(VecEnv):
    """
    We fork the subprocessing from the stable-baselines implementation, but use RaggedBuffers for collecting batches

    Citation here: https://github.com/DLR-RM/stable-baselines3/blob/master/CITATION.bib
    """

    def __init__(
        self,
        create_env: Callable[[], Environment],
        num_envs: int,
        num_processes: int,
        start_method: Optional[str] = None,
    ):

        if start_method is None:
            # Fork is not a thread safe method (see issue #217)
            # but is more user friendly (does not require to wrap the code in
            # a `if __name__ == "__main__":`)
            forkserver_available = "forkserver" in mp.get_all_start_methods()
            start_method = "forkserver" if forkserver_available else "spawn"
        ctx = mp.get_context(start_method)

        assert (
            num_envs % num_processes == 0
        ), "The required number of environments can not be equally split into the number of specified processes."

        self.num_processes = num_processes
        self.num_envs = num_envs
        self.envs_per_process = int(num_envs / num_processes)

        env_list_configs = [
            (create_env, self.envs_per_process) for _ in range(self.num_processes)
        ]

        self.remotes = []
        self.work_remotes = []
        for i in range(self.num_processes):
            pipe = ctx.Pipe()
            self.remotes.append(MsgpackConnectionWrapper(pipe[0]))
            self.work_remotes.append(MsgpackConnectionWrapper(pipe[1]))

        self.processes = []
        for work_remote, remote, env_list_config in zip(
            self.work_remotes, self.remotes, env_list_configs
        ):
            # Have to use cloudpickle wrapper here to serialize the ABCMeta class reference
            # TODO: Can this be achieved with custom msgpack somehow?
            args = (work_remote, remote, CloudpickleWrapper(env_list_config))
            # daemon=True: if the main process crashes, we should not cause things to hang
            process = ctx.Process(
                target=_worker, args=args, daemon=True
            )  # pytype:disable=attribute-error
            process.start()
            self.processes.append(process)
            work_remote.close()

        env = create_env()
        self._obs_space = env.obs_space()
        self._action_space = env.action_space()

    def reset(self, obs_space: ObsSpace) -> VecObs:
        for remote in self.remotes:
            remote.send(("reset", obs_space))

        # Empty initialized observation batch
        observations = batch_obs([], self.obs_space(), self.action_space())

        for remote in self.remotes:
            remote_obs_batch = remote.recv()
            observations.extend(remote_obs_batch)

        assert isinstance(observations, VecObs)
        return observations

    def render(self, **kwargs: Any) -> npt.NDArray[np.uint8]:
        rgb_arrays = []
        for remote in self.remotes:
            remote.send(("render", kwargs))
            rgb_arrays.append(remote.recv())

        np_rgb_arrays = np.concatenate(rgb_arrays)
        assert isinstance(np_rgb_arrays, np.ndarray)
        return np_rgb_arrays

    def close(self) -> None:
        for remote in self.remotes:
            remote.send(("close", None))
        for process in self.processes:
            process.join()

    def _chunk_actions(
        self, actions: Mapping[str, RaggedBufferI64]
    ) -> Generator[Mapping[str, RaggedBufferI64], List[Observation], None]:
        for i in range(0, self.num_envs, self.envs_per_process):
            yield {
                atype: a[i : i + self.envs_per_process, :, :]
                for atype, a in actions.items()
            }

    def act(
        self, actions: Mapping[str, RaggedBufferI64], obs_space: ObsSpace
    ) -> VecObs:
        remote_actions = self._chunk_actions(actions)
        for remote, action in zip(self.remotes, remote_actions):
            remote.send(("act", (action, obs_space)))

        # Empty initialized observation batch
        observations = batch_obs([], self.obs_space(), self.action_space())

        for remote in self.remotes:
            remote_obs_batch = remote.recv()
            observations.extend(remote_obs_batch)
        return observations

    def __len__(self) -> int:
        return self.num_envs

    def obs_space(self) -> ObsSpace:
        return self._obs_space

    def action_space(self) -> Dict[ActionName, ActionSpace]:
        return self._action_space



import random
from dataclasses import dataclass
from typing import Dict, Mapping

import numpy as np

from entity_gym.env import (
    Action,
    ActionSpace,
    CategoricalAction,
    CategoricalActionMask,
    CategoricalActionSpace,
    Entity,
    Environment,
    Observation,
    ObsSpace,
)


@dataclass
class NotHotdog(Environment):
    """
    On each timestep, there is either a generic "Object" entity with a `is_hotdog` property, or a "Hotdog" object.
    The "Player" entity is always present, and has an action to classify the other entity as hotdog or not hotdog.
    """

    def obs_space(self) -> ObsSpace:
        return ObsSpace(
            entities={
                "Player": Entity(["step"]),
                "Object": Entity(["is_hotdog"]),
                "Hotdog": Entity([]),
            }
        )

    def action_space(self) -> Dict[str, ActionSpace]:
        return {
            "classify": CategoricalActionSpace(["hotdog", "not_hotdog"]),
            "unused_action": CategoricalActionSpace(["0", "1"]),
        }

    def reset(self) -> Observation:
        self.step = 0
        self.is_hotdog = random.randint(0, 1)
        self.hotdog_object = random.randint(0, 1) == 1
        return self.observe()

    def act(self, actions: Mapping[str, Action]) -> Observation:
        self.step += 1

        a = actions["classify"]
        assert isinstance(a, CategoricalAction), f"{a} is not a CategoricalAction"
        if a.indices[0] == self.is_hotdog:
            reward = 1
        else:
            reward = 0
        done = True
        return self.observe(done, reward)

    def observe(self, done: bool = False, reward: float = 0) -> Observation:
        return Observation(
            features={
                "Player": np.array(
                    [
                        [
                            self.step,
                        ]
                    ],
                    dtype=np.float32,
                ),
                "Object": np.array(
                    [
                        [
                            self.is_hotdog,
                        ]
                    ],
                    dtype=np.float32,
                )
                if (self.hotdog_object and self.is_hotdog == 0)
                or not self.hotdog_object
                else np.zeros((0, 1), dtype=np.float32).reshape(0, 1),
                "Hotdog": np.zeros((1, 0), dtype=np.float32)
                if self.hotdog_object and self.is_hotdog == 1
                else np.zeros((0, 0), dtype=np.float32),
            },
            actions={
                "classify": CategoricalActionMask(actor_ids=[0]),
                "unused_action": CategoricalActionMask(actor_ids=[]),
            },
            ids={"Player": [0]},
            reward=reward,
            done=done,
        )



import random
from dataclasses import dataclass, field
from typing import Dict, List, Mapping

import numpy as np

from entity_gym.env import (
    Action,
    ActionSpace,
    Entity,
    Environment,
    Observation,
    ObsSpace,
    SelectEntityAction,
    SelectEntityActionMask,
    SelectEntityActionSpace,
)
from entity_gym.env.environment import ActionName


@dataclass
class Ball:
    color: int
    selected: bool = False


@dataclass
class PickMatchingBalls(Environment):
    """
    The PickMatchingBalls environment is initialized with a list of 32 balls of different colors.
    On each timestamp, the player can pick up one of the balls.
    The episode ends when the player picks up a ball of a different color from the last one.
    The player receives a reward equal to the number of balls picked up divided by the maximum number of balls of the same color.
    """

    max_balls: int = 32
    balls: List[Ball] = field(default_factory=list)
    one_hot: bool = False  # use one-hot encoding for the ball color feature
    randomize: bool = (
        False  # randomize the number of balls to be between 3 and max_balls
    )

    def obs_space(self) -> ObsSpace:
        return ObsSpace(
            entities={
                "Ball": Entity(
                    # TODO: better support for categorical features
                    [
                        "color0",
                        "color1",
                        "color2",
                        "color3",
                        "color4",
                        "color5",
                        "selected",
                    ],
                ),
                "Player": Entity([]),
            }
        )

    def action_space(self) -> Dict[ActionName, ActionSpace]:
        return {"Pick Ball": SelectEntityActionSpace()}

    def reset(self) -> Observation:
        num_balls = (
            self.max_balls if not self.randomize else random.randint(3, self.max_balls)
        )
        self.balls = [Ball(color=random.randint(0, 5)) for _ in range(num_balls)]
        return self.observe()

    def observe(self) -> Observation:
        done = len({b.color for b in self.balls if b.selected}) > 1 or all(
            b.selected for b in self.balls
        )
        if done:
            if all(b.selected for b in self.balls):
                reward = 1.0
            else:
                reward = (sum(b.selected for b in self.balls) - 1) / max(
                    len([b for b in self.balls if b.color == color])
                    for color in range(6)
                )
        else:
            reward = 0.0

        return Observation(
            features={
                "Ball": np.array(
                    [
                        [float(b.color == c) for c in range(6)] + [float(b.selected)]
                        for b in self.balls
                    ]
                    if self.one_hot
                    else [
                        [float(b.color) for _ in range(6)] + [float(b.selected)]
                        for b in self.balls
                    ],
                    dtype=np.float32,
                ),
                "Player": np.zeros([1, 0], dtype=np.float32),
            },
            ids={
                "Ball": list(range(len(self.balls))),
                "Player": [len(self.balls)],
            },
            actions={
                "Pick Ball": SelectEntityActionMask(
                    actor_ids=[len(self.balls)],
                    actee_ids=[i for i, b in enumerate(self.balls) if not b.selected],
                ),
            },
            reward=reward,
            done=done,
        )

    def act(self, actions: Mapping[ActionName, Action]) -> Observation:
        action = actions["Pick Ball"]
        assert isinstance(action, SelectEntityAction)
        for selected_ball in action.actees:
            assert not self.balls[selected_ball].selected
            self.balls[selected_ball].selected = True
        return self.observe()



from typing import Any, Dict, Mapping, Optional

import numpy as np
import numpy.typing as npt
from ragged_buffer import RaggedBufferI64

from entity_gym.env.environment import ActionName, ActionSpace, ObsSpace
from entity_gym.env.vec_env import Metric, VecEnv, VecObs


class AddMetricsWrapper(VecEnv):
    def __init__(
        self, env: VecEnv, filter: Optional[npt.NDArray[np.bool8]] = None
    ) -> None:
        """
        Wrap a VecEnv to track and add metrics for (episodic) rewards and episode lengths.

        Args:
            env: The VecEnv to wrap.
            filter: A boolean array of length len(env) indicating which environments to
                track metrics for. If filter[i] is True, then the metrics for the i-th
                environment will be tracked.
        """
        self.env = env
        self.total_reward = np.zeros(len(env), dtype=np.float32)
        self.total_steps = np.zeros(len(env), dtype=np.int64)
        self.filter = np.ones(len(env), dtype=np.bool8) if filter is None else filter

    def reset(self, obs_config: ObsSpace) -> VecObs:
        return self.track_metrics(self.env.reset(obs_config))

    def act(
        self, actions: Mapping[ActionName, RaggedBufferI64], obs_filter: ObsSpace
    ) -> VecObs:
        return self.track_metrics(self.env.act(actions, obs_filter))

    def render(self, **kwargs: Any) -> npt.NDArray[np.uint8]:
        return self.env.render(**kwargs)

    def __len__(self) -> int:
        return len(self.env)

    def close(self) -> None:
        self.env.close()

    def track_metrics(self, obs: VecObs) -> VecObs:
        self.total_reward += obs.reward
        self.total_steps += 1
        episodic_reward = Metric()
        episodic_length = Metric()
        obs.metrics["step"] = Metric(
            sum=self.total_steps.sum(),
            count=len(self.total_steps),
            min=self.total_steps.min(),
            max=self.total_steps.max(),
        )
        for i in np.arange(len(self))[obs.done & self.filter]:
            episodic_reward.push(self.total_reward[i])
            episodic_length.push(self.total_steps[i])
            self.total_reward[i] = 0.0
            self.total_steps[i] = 0
        obs.metrics["episodic_reward"] = episodic_reward
        obs.metrics["episode_length"] = episodic_length
        obs.metrics["reward"] = Metric(
            sum=obs.reward.sum(),
            count=obs.reward.size,
            min=obs.reward.min(),
            max=obs.reward.max(),
        )
        return obs

    def action_space(self) -> Dict[ActionName, ActionSpace]:
        return self.env.action_space()

    def obs_space(self) -> ObsSpace:
        return self.env.obs_space()



from abc import ABC, abstractmethod
from typing import Dict, Optional, Tuple

import click
import numpy as np

from entity_gym.env import (
    Action,
    CategoricalAction,
    CategoricalActionSpace,
    Environment,
    Observation,
    ObsSpace,
    SelectEntityAction,
    SelectEntityActionMask,
    SelectEntityActionSpace,
)
from entity_gym.env.environment import (
    Action,
    GlobalCategoricalAction,
    GlobalCategoricalActionMask,
    GlobalCategoricalActionSpace,
    Observation,
)
from entity_gym.env.validator import ValidatingEnv


class Agent(ABC):
    """Interface for an agent that receives observations and outputs actions."""

    @abstractmethod
    def act(self, obs: Observation) -> Tuple[Dict[str, Action], float]:
        pass


class CliRunner:
    """
    Interactively run any entity gym environment in a CLI.

    Example:

    .. code-block:: pycon

        >>> from entity_gym.runner import CliRunner
        >>> from entity_gym.examples import TreasureHunt
        >>> CliRunner(TreasureHunt()).run()
    """

    def __init__(self, env: Environment, agent: Optional[Agent] = None) -> None:
        self.env = ValidatingEnv(env)
        self.agent = agent

    def run(self, restart: bool = False) -> None:
        print_env(self.env)
        print()

        obs_space = self.env.obs_space()
        actions = self.env.action_space()
        obs = self.env.reset_filter(obs_space)
        total_reward = obs.reward
        step = 0
        while True:
            if obs.done:
                if restart:
                    obs = self.env.reset_filter(obs_space)
                    total_reward = obs.reward
                    step = 0
                else:
                    break
            if self.agent is None:
                agent_action: Optional[Dict[str, Action]] = None
                agent_prediction: Optional[float] = None
            else:
                agent_action, agent_prediction = self.agent.act(obs)
            print_obs(step, obs, total_reward, obs_space, agent_prediction)
            action: Dict[str, Action] = {}
            received_action = False
            for action_name, action_mask in obs.actions.items():
                action_def = actions[action_name]
                if isinstance(action_mask, GlobalCategoricalActionMask):
                    assert isinstance(action_def, GlobalCategoricalActionSpace)
                    if agent_action is None:
                        choices = " ".join(
                            f"{i}/{label}"
                            for i, label in enumerate(action_def.index_to_label)
                        )
                    else:
                        probs = agent_action[action_name].probs
                        assert probs is not None
                        choice_id = agent_action[action_name].index  # type: ignore
                        choices = " ".join(
                            click.style(
                                f"{i}/{label} ",
                                fg="yellow" if i == choice_id else None,
                                bold=i == choice_id,
                            )
                            + click.style(f"{100 * prob:.1f}%", fg="yellow")
                            for i, (label, prob) in enumerate(
                                zip(action_def.index_to_label, probs)
                            )
                        )

                    click.echo(
                        f"Choose "
                        + click.style(f"{action_name}", fg="green")
                        + f" ({choices})"
                    )
                    try:
                        inp = input()
                        if inp == "" and agent_action is not None:
                            choice_id = agent_action[action_name].index  # type: ignore
                        else:
                            choice_id = int(inp)
                        received_action = True
                    except KeyboardInterrupt:
                        print()
                        print("Exiting")
                        return
                    action[action_name] = GlobalCategoricalAction(
                        index=choice_id,
                        label=action_def.index_to_label[choice_id],
                    )
                    continue
                elif action_mask.actor_ids is not None:
                    actor_ids = action_mask.actor_ids
                elif action_mask.actor_types is not None:
                    actor_ids = [
                        id for atype in action_mask.actor_types for id in obs.ids[atype]
                    ]
                else:
                    actor_ids = obs.index_to_id(obs_space)

                print()
                for actor_id in actor_ids:
                    if isinstance(action_def, CategoricalActionSpace):
                        # Prompt user for action
                        if agent_action is None:
                            choices = " ".join(
                                f"{i}/{label}"
                                for i, label in enumerate(action_def.index_to_label)
                            )
                        else:
                            aa = agent_action[action_name]
                            assert isinstance(aa, CategoricalAction)
                            actor_index = aa.actors.index(actor_id)
                            probs = aa.probs[actor_index]  # type: ignore
                            assert probs is not None
                            choice_id = aa.indices[actor_index]
                            choices = " ".join(
                                click.style(
                                    f"{i}/{label} ",
                                    fg="yellow" if i == choice_id else None,
                                    bold=i == choice_id,
                                )
                                + click.style(f"{100 * prob:.1f}%", fg="yellow")
                                for i, (label, prob) in enumerate(
                                    zip(action_def.index_to_label, probs)
                                )
                            )
                        click.echo(
                            f"Choose "
                            + click.style(f"{action_name}", fg="green")
                            + f" for actor {actor_id}"
                            + f" ({choices})"
                        )

                        try:
                            inp = input()
                            if inp == "" and agent_action is not None:
                                aa = agent_action[action_name]
                                assert isinstance(aa, CategoricalAction)
                                choice_id = aa.indices[actor_index]
                            else:
                                choice_id = int(inp)
                            received_action = True
                        except KeyboardInterrupt:
                            print()
                            print("Exiting")
                            return
                        if action_name not in action:
                            action[action_name] = CategoricalAction(
                                indices=np.zeros(
                                    (0, len(action_def.index_to_label)), dtype=np.int64
                                ),
                                index_to_label=action_def.index_to_label,
                                actors=[],
                            )
                        a = action[action_name]
                        assert isinstance(a, CategoricalAction)
                        a.indices = np.array(list(a.indices) + [choice_id])
                        a.actors = list(a.actors) + [actor_id]
                    elif isinstance(action_def, SelectEntityActionSpace):
                        assert isinstance(action_mask, SelectEntityActionMask)
                        # Prompt user for entity
                        click.echo(
                            f"Choose "
                            + click.style(f"{action_name}", fg="green")
                            + f" for actor {actor_id}"
                        )
                        if action_mask.actee_ids is not None:
                            print(
                                f"Selectable entities: {', '.join([str(id) for id in action_mask.actee_ids])}"
                            )
                        elif action_mask.actee_types is not None:
                            print(
                                f"Selectable entity types: {', '.join([str(id) for id in action_mask.actee_types])}"
                            )
                        else:
                            print("Selectable entities: all")

                        try:
                            entity_id = int(input())
                        except KeyboardInterrupt:
                            print()
                            print("Exiting")
                            return
                        received_action = True
                        if action_name not in action:
                            action[action_name] = SelectEntityAction([], [])
                        a = action[action_name]
                        assert isinstance(a, SelectEntityAction)
                        a.actors = list(a.actors) + [actor_id]
                        a.actees = list(a.actees) + [entity_id]
                    else:
                        raise ValueError(f"Unknown action type {action_def}")
            if not received_action:
                try:
                    input("Press ENTER to continue, CTRL-C to exit")
                except KeyboardInterrupt:
                    print()
                    print("Exiting")
                    return
            obs = self.env.act_filter(action, obs_space)
            total_reward += obs.reward
            step += 1

        print_obs(step, obs, total_reward, obs_space)
        click.secho("Episode finished", fg="green")


def print_env(env: ValidatingEnv) -> None:
    click.secho(f"Environment: {env.env.__class__.__name__}", fg="white", bold=True)
    obs = env.obs_space()
    if len(obs.global_features) > 0:
        click.echo(
            click.style("Global features: ", fg="cyan") + ", ".join(obs.global_features)
        )
    for label, entity in obs.entities.items():
        click.echo(
            click.style(f"Entity ", fg="cyan")
            + click.style(f"{label}", fg="green")
            + click.style(f": " if len(entity.features) > 0 else "", fg="cyan")
            + ", ".join(entity.features)
        )
    acts = env.action_space()
    for label, action in acts.items():
        if isinstance(action, CategoricalActionSpace) or isinstance(
            action, GlobalCategoricalActionSpace
        ):
            click.echo(
                click.style(f"Categorical", fg="cyan")
                + click.style(f" {label}", fg="green")
                + click.style(f": ", fg="cyan")
                + ", ".join(action.index_to_label)
            )
        elif isinstance(action, SelectEntityActionSpace):
            click.echo(
                click.style(f"Select entity", fg="cyan")
                + click.style(f" {label}", fg="green")
            )
        else:
            raise ValueError(f"Unknown action type {action}")


def print_obs(
    step: int,
    obs: Observation,
    total_reward: float,
    obs_filter: ObsSpace,
    predicted_return: Optional[float] = None,
) -> None:
    click.secho(f"Step {step}", fg="white", bold=True)
    click.echo(click.style("Reward: ", fg="cyan") + f"{obs.reward}")
    click.echo(click.style("Total: ", fg="cyan") + f"{total_reward}")
    if predicted_return is not None:
        click.echo(
            click.style("Predicted return: ", fg="cyan")
            + click.style(f"{predicted_return:.3e}", fg="yellow")
        )
    if len(obs_filter.global_features) > 0:
        click.echo(
            click.style("Global features: ", fg="cyan")
            + ", ".join(
                f"{label}={value}"
                for label, value in zip(obs_filter.global_features, obs.global_features)
            )
        )
    if len(obs_filter.entities) > 0:
        click.echo(click.style("Entities", fg="cyan"))
    entity_index = 0
    for entity_type, features in obs.features.items():
        for entity in range(len(features)):
            if entity_type in obs.ids:
                id = f" (id={obs.ids[entity_type][entity]})"
            else:
                id = ""
            rendered = (
                click.style(entity_type, fg="green")
                + "("
                + ", ".join(
                    map(
                        lambda nv: nv[0] + "=" + str(nv[1]),
                        zip(
                            obs_filter.entities[entity_type].features,
                            features[entity],
                        ),
                    )
                )
                + ")"
            )
            print(f"{entity_index} {rendered}{id}")
            entity_index += 1


__all__ = ["CliRunner", "Agent"]



from dataclasses import dataclass
from typing import Dict, Mapping

from entity_gym.env import Action, ActionSpace, Environment, Observation, ObsSpace
from entity_gym.env.environment import (
    GlobalCategoricalAction,
    GlobalCategoricalActionMask,
    GlobalCategoricalActionSpace,
)


@dataclass
class MultiArmedBandit(Environment):
    """
    Task with single cateorical action with 5 choices which gives a reward of 1 for choosing action 0 and reward of 0 otherwise.
    """

    def obs_space(cls) -> ObsSpace:
        return ObsSpace(global_features=["step"])

    def action_space(cls) -> Dict[str, ActionSpace]:
        return {
            "pull": GlobalCategoricalActionSpace(["A", "B", "C", "D", "E"]),
        }

    def reset(self) -> Observation:
        self.step = 0
        self._total_reward = 0.0
        return self.observe()

    def act(self, actions: Mapping[str, Action]) -> Observation:
        self.step += 1

        a = actions["pull"]
        assert isinstance(
            a, GlobalCategoricalAction
        ), f"{a} is not a GlobalCategoricalAction"
        if a.label == "A":
            reward = 1 / 32.0
        else:
            reward = 0
        done = self.step >= 32
        self._total_reward += reward
        return self.observe(done, reward)

    def observe(self, done: bool = False, reward: float = 0) -> Observation:
        return Observation(
            global_features=[self.step],
            actions={
                "pull": GlobalCategoricalActionMask(),
            },
            reward=reward,
            done=done,
        )



from typing import Any, Sequence, Union

import numpy as np
import numpy.typing as npt

Features = Union[npt.NDArray[np.float32], Sequence[Sequence[float]]]
EntityID = Any
EntityName = str
ActionName = str



import random
from typing import Dict, List, Mapping, Tuple

from entity_gym.env import *


class MineSweeper(Environment):
    """
    The MineSweeper environment contains two types of objects, mines and robots.
    The player controls all robots in the environment.
    On every step, each robot may move in one of four cardinal directions, or stay in place and defuse all adjacent mines.
    If a robot defuses a mine, it is removed from the environment.
    If a robot steps on a mine, it is removed from the environment and the player loses the game.
    The player wins the game when all mines are defused.
    """

    def __init__(
        self,
        width: int = 6,
        height: int = 6,
        nmines: int = 5,
        nrobots: int = 2,
        orbital_cannon: bool = False,
        cooldown_period: int = 5,
    ):
        self.width = width
        self.height = height
        self.nmines = nmines
        self.nrobots = nrobots
        self.orbital_cannon = orbital_cannon
        self.cooldown_period = cooldown_period
        self.orbital_cannon_cooldown = cooldown_period
        # Positions of robots and mines
        self.robots: List[Tuple[int, int]] = []
        self.mines: List[Tuple[int, int]] = []

    def obs_space(cls) -> ObsSpace:
        return ObsSpace(
            entities={
                "Mine": Entity(features=["x", "y"]),
                "Robot": Entity(features=["x", "y"]),
                "Orbital Cannon": Entity(["cooldown"]),
            }
        )

    def action_space(cls) -> Dict[ActionName, ActionSpace]:
        return {
            "Move": CategoricalActionSpace(
                ["Up", "Down", "Left", "Right", "Defuse Mines"],
            ),
            "Fire Orbital Cannon": SelectEntityActionSpace(),
        }

    def reset(self) -> Observation:
        positions = random.sample(
            [(x, y) for x in range(self.width) for y in range(self.height)],
            self.nmines + self.nrobots,
        )
        self.mines = positions[: self.nmines]
        self.robots = positions[self.nmines :]
        self.orbital_cannon_cooldown = self.cooldown_period
        return self.observe()

    def observe(self) -> Observation:
        done = len(self.mines) == 0 or len(self.robots) == 0
        reward = 1.0 if len(self.mines) == 0 else 0.0
        return Observation(
            entities={
                "Mine": (
                    self.mines,
                    [("Mine", i) for i in range(len(self.mines))],
                ),
                "Robot": (
                    self.robots,
                    [("Robot", i) for i in range(len(self.robots))],
                ),
                "Orbital Cannon": (
                    [(self.orbital_cannon_cooldown,)],
                    [("Orbital Cannon", 0)],
                )
                if self.orbital_cannon
                else None,
            },
            actions={
                "Move": CategoricalActionMask(
                    # Allow all robots to move
                    actor_types=["Robot"],
                    mask=[self.valid_moves(x, y) for x, y in self.robots],
                ),
                "Fire Orbital Cannon": SelectEntityActionMask(
                    # Only the Orbital Cannon can fire, but not if cooldown > 0
                    actor_types=["Orbital Cannon"]
                    if self.orbital_cannon_cooldown == 0
                    else [],
                    # Both mines and robots can be fired at
                    actee_types=["Mine", "Robot"],
                ),
            },
            # The game is done once there are no more mines or robots
            done=done,
            # Give reward of 1.0 for defusing all mines
            reward=reward,
        )

    def act(self, actions: Mapping[ActionName, Action]) -> Observation:
        fire = actions["Fire Orbital Cannon"]
        assert isinstance(fire, SelectEntityAction)
        remove_robot = None
        for (entity_type, i) in fire.actees:
            if entity_type == "Mine":
                self.mines.remove(self.mines[i])
            elif entity_type == "Robot":
                # Don't remove yet to keep indices valid
                remove_robot = i

        move = actions["Move"]
        assert isinstance(move, CategoricalAction)
        for (_, i), choice in zip(move.actors, move.indices):
            if self.robots[i] is None:
                continue
            # Action space is ["Up", "Down", "Left", "Right", "Defuse Mines"],
            x, y = self.robots[i]
            if choice == 0 and y < self.height - 1:
                self.robots[i] = (x, y + 1)
            elif choice == 1 and y > 0:
                self.robots[i] = (x, y - 1)
            elif choice == 2 and x > 0:
                self.robots[i] = (x - 1, y)
            elif choice == 3 and x < self.width - 1:
                self.robots[i] = (x + 1, y)
            elif choice == 4:
                # Remove all mines adjacent to this robot
                rx, ry = self.robots[i]
                self.mines = [
                    (x, y) for (x, y) in self.mines if abs(x - rx) + abs(y - ry) > 1
                ]

        if remove_robot is not None:
            self.robots.pop(remove_robot)
        # Remove all robots that stepped on a mine
        self.robots = [r for r in self.robots if r not in self.mines]

        return self.observe()

    def valid_moves(self, x: int, y: int) -> List[bool]:
        return [
            x < self.width - 1,
            x > 0,
            y < self.height - 1,
            y > 0,
            # Always allow staying in place and defusing mines
            True,
        ]



from typing import Dict, Type

from entity_gym.env import Environment
from entity_gym.examples.cherry_pick import CherryPick
from entity_gym.examples.count import Count
from entity_gym.examples.floor_is_lava import FloorIsLava
from entity_gym.examples.minefield import Minefield
from entity_gym.examples.minesweeper import MineSweeper
from entity_gym.examples.move_to_origin import MoveToOrigin
from entity_gym.examples.multi_armed_bandit import MultiArmedBandit
from entity_gym.examples.multi_snake import MultiSnake
from entity_gym.examples.not_hotdog import NotHotdog
from entity_gym.examples.pick_matching_balls import PickMatchingBalls
from entity_gym.examples.rock_paper_scissors import RockPaperScissors
from entity_gym.examples.tutorial import TreasureHunt
from entity_gym.examples.xor import Xor

ENV_REGISTRY: Dict[str, Type[Environment]] = {
    "MoveToOrigin": MoveToOrigin,
    "CherryPick": CherryPick,
    "PickMatchingBalls": PickMatchingBalls,
    "Minefield": Minefield,
    "MultiSnake": MultiSnake,
    "MultiArmedBandit": MultiArmedBandit,
    "NotHotdog": NotHotdog,
    "Xor": Xor,
    "Count": Count,
    "FloorIsLava": FloorIsLava,
    "MineSweeper": MineSweeper,
    "RockPaperScissors": RockPaperScissors,
    "TreasureHunt": TreasureHunt,
}

__all__ = [
    "MoveToOrigin",
    "CherryPick",
    "PickMatchingBalls",
    "Minefield",
    "MultiSnake",
    "MultiArmedBandit",
    "NotHotdog",
    "Xor",
    "Count",
    "FloorIsLava",
    "MineSweeper",
    "RockPaperScissors",
    "TreasureHunt",
]



import time
from collections import defaultdict
from contextlib import contextmanager
from typing import DefaultDict, Dict, Generator, List


class Tracer:
    def __init__(self, cuda: bool = True) -> None:
        if cuda:
            try:
                import torch

                cuda = torch.cuda.is_available()
            except ImportError:
                cuda = False

        self.start_time: List[float] = []
        self.callstack: List[str] = []
        self.total_time: DefaultDict[str, float] = defaultdict(float)
        self.cuda = cuda

    def start(self, name: str) -> None:
        self.callstack.append(name)
        self.start_time.append(time.time())

    def end(self, name: str) -> None:
        if self.cuda:
            import torch

            torch.cuda.synchronize()
        self.total_time[self.stack] += time.time() - self.start_time.pop()
        actual_name = self.callstack.pop()
        assert (
            actual_name == name
        ), f"Expected to complete {name}, but currently active span is {actual_name}"

    def finish(self) -> Dict[str, float]:
        assert (
            len(self.callstack) == 0
        ), f"Cannot finish when there are open traces: {self.stack}"
        self_times: Dict[str, float] = {}
        # Traverse the tree depth-first
        for name in reversed(sorted(self.total_time.keys())):
            time_in_children = sum(
                t for child, t in self_times.items() if child.startswith(name)
            )
            self_times[f"{name}[self]"] = self.total_time[name] - time_in_children

        self_times.update(self.total_time)
        self.total_time = defaultdict(float)
        return self_times

    @contextmanager
    def span(self, name: str) -> Generator[None, None, None]:
        self.start(name)
        yield
        self.end(name)

    @property
    def stack(self) -> str:
        return ".".join(self.callstack)



import warnings
from typing import Any

import msgpack_numpy
from ragged_buffer import RaggedBufferBool, RaggedBufferF32, RaggedBufferI64

from entity_gym.env.environment import *
from entity_gym.env.vec_env import *

# For security reasons we don't want to deserialize classes that are not in this list.
WHITELIST = {
    "ObsSpace": ObsSpace,
    "VecObs": VecObs,
    "VecCategoricalActionMask": VecCategoricalActionMask,
    "VecSelectEntityActionMask": VecSelectEntityActionMask,
    "SelectEntityAction": SelectEntityAction,
    "CategoricalAction": CategoricalAction,
    "GlobalCategoricalAction": GlobalCategoricalAction,
    "SelectEntityActionSpace": SelectEntityActionSpace,
    "CategoricalActionSpace": CategoricalActionSpace,
    "GlobalCategoricalActionSpace": GlobalCategoricalActionSpace,
    "Entity": Entity,
    "Metric": Metric,
}


def ragged_buffer_encode(obj: Any) -> Any:
    # Suppress "msgpack_numpy.py:96: FutureWarning: elementwise comparison failed; returning scalar instead, but in the future will perform elementwise comparison"
    with warnings.catch_warnings():
        warnings.simplefilter(action="ignore", category=FutureWarning)

        if isinstance(obj, RaggedBufferF32) or isinstance(obj, RaggedBufferI64) or isinstance(obj, RaggedBufferBool):  # type: ignore
            flattened = obj.as_array()
            lengths = obj.size1()
            return {
                "__flattened__": msgpack_numpy.encode(flattened),
                "__lengths__": msgpack_numpy.encode(lengths),
            }
        elif hasattr(obj, "__dict__"):
            return {"__classname__": obj.__class__.__name__, "data": vars(obj)}
        else:
            return obj


def ragged_buffer_decode(obj: Any) -> Any:
    # Suppress "msgpack_numpy.py:96: FutureWarning: elementwise comparison failed; returning scalar instead, but in the future will perform elementwise comparison"
    with warnings.catch_warnings():
        warnings.simplefilter(action="ignore", category=FutureWarning)

        if "__flattened__" in obj:
            flattened = msgpack_numpy.decode(obj["__flattened__"])
            lengths = msgpack_numpy.decode(obj["__lengths__"])

            dtype = flattened.dtype

            if dtype == np.float32:
                return RaggedBufferF32.from_flattened(flattened, lengths)
            elif dtype == int:
                return RaggedBufferI64.from_flattened(flattened, lengths)
            elif dtype == bool:
                return RaggedBufferBool.from_flattened(flattened, lengths)
            else:
                raise ValueError(f"Unsupported RaggedBuffer dtype: {dtype}")
        elif "__classname__" in obj:
            classname = obj["__classname__"]
            if classname in WHITELIST:
                cls_name = globals()[classname]
                return cls_name(**obj["data"])
            else:
                raise RuntimeError(
                    f"Attempt to deserialize class {classname} outside whitelist."
                )
        else:
            return obj



import argparse

from entity_gym.examples import ENV_REGISTRY
from entity_gym.runner import CliRunner

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--env", type=str, default="MoveToOrigin")
    args = parser.parse_args()

    envs = ENV_REGISTRY
    if args.env not in envs:
        raise ValueError(
            f"Unknown environment {args.env}\nValid environments are {list(envs.keys())}"
        )
    else:
        env_cls = envs[args.env]

    print(env_cls)
    env = env_cls()
    CliRunner(env).run()



import random
from dataclasses import dataclass
from typing import Dict, Mapping

from entity_gym.dataclass_utils import extract_features, obs_space_from_dataclasses
from entity_gym.env import (
    Action,
    ActionSpace,
    CategoricalAction,
    CategoricalActionMask,
    CategoricalActionSpace,
    Environment,
    Observation,
    ObsSpace,
)


@dataclass
class Player:
    pass


@dataclass
class Opponent:
    rock: float = 0.0
    paper: float = 0.0
    scissors: float = 0.0


class RockPaperScissors(Environment):
    """
    This environment tests giving additional information to the value function
    which can not be observed by the policy.

    On each timestep, the opponent randomly chooses rock, paper or scissors with
    probability of 50%, 30% and 20% respectively. The value function can observe
    the opponent's choice, but the policy can not.
    The agent must choose either rock, paper or scissors. If the agent beats the
    opponent, the agent receives a reward of 2.0, otherwise it receives a reward of 0.0.
    The optimal strategy is to always choose paper for an average reward of 1.0.
    Since the value function can observe the opponent's choice, it can perfectly
    predict reward.
    """

    def __init__(self, cheat: bool = False) -> None:
        self.cheat = cheat
        self.reset()

    def obs_space(self) -> ObsSpace:
        return obs_space_from_dataclasses(Player, Opponent)

    def action_space(self) -> Dict[str, ActionSpace]:
        return {"throw": CategoricalActionSpace(["rock", "paper", "scissors"])}

    def reset_filter(self, obs_space: ObsSpace) -> Observation:
        rand = random.random()
        if rand < 0.5:
            self.opponent = Opponent(rock=1.0)
        elif rand < 0.8:
            self.opponent = Opponent(paper=1.0)
        else:
            self.opponent = Opponent(scissors=1.0)
        return self.observe(obs_space)

    def reset(self) -> Observation:
        return self.reset_filter(self.obs_space())

    def act_filter(
        self, action: Mapping[str, Action], obs_filter: ObsSpace
    ) -> Observation:
        reward = 0.0
        for action_name, a in action.items():
            assert isinstance(a, CategoricalAction)
            if action_name == "throw":
                if (
                    (a.indices[0] == 0 and self.opponent.scissors == 1.0)
                    or (a.indices[0] == 1 and self.opponent.rock == 1.0)
                    or (a.indices[0] == 2 and self.opponent.paper == 1.0)
                ):
                    reward = 2.0
        return self.observe(obs_filter, done=True, reward=reward)

    def act(self, actions: Mapping[str, Action]) -> Observation:
        return self.act_filter(
            actions,
            self.obs_space(),
        )

    def observe(
        self, obs_filter: ObsSpace, done: bool = False, reward: float = 0.0
    ) -> Observation:
        return Observation(
            features=extract_features(
                {
                    "Player": [Player()],
                    "Opponent": [self.opponent],
                },
                obs_filter,
            ),
            actions={
                "throw": CategoricalActionMask(actor_ids=[0]),
            },
            ids={"Player": [0]},
            visible={"Opponent": [self.cheat]},
            reward=reward,
            done=done,
        )



from typing import Any, Callable, Dict, List, Mapping, Optional

import numpy as np
import numpy.typing as npt
from ragged_buffer import RaggedBufferI64

from entity_gym.env.environment import (
    Action,
    ActionName,
    ActionSpace,
    CategoricalAction,
    CategoricalActionMask,
    CategoricalActionSpace,
    EntityID,
    Environment,
    GlobalCategoricalAction,
    GlobalCategoricalActionSpace,
    Observation,
    ObsSpace,
    SelectEntityAction,
    SelectEntityActionMask,
    SelectEntityActionSpace,
)
from entity_gym.env.vec_env import VecEnv, VecObs, batch_obs


class EnvList(VecEnv):
    def __init__(self, create_env: Callable[[], Environment], num_envs: int):
        self.envs = [create_env() for _ in range(num_envs)]
        self.last_obs: List[Observation] = []
        env = self.envs[0] if num_envs > 0 else create_env()
        self._obs_space = env.obs_space()
        self._action_space = env.action_space()

    def reset(self, obs_space: ObsSpace) -> VecObs:
        batch = self._batch_obs([e.reset_filter(obs_space) for e in self.envs])
        return batch

    def render(self, **kwargs: Any) -> npt.NDArray[np.uint8]:
        return np.stack([e.render(**kwargs) for e in self.envs])

    def close(self) -> None:
        for env in self.envs:
            env.close()

    def act(
        self, actions: Mapping[str, RaggedBufferI64], obs_space: ObsSpace
    ) -> VecObs:
        observations = []
        action_spaces = self.action_space()
        for i, env in enumerate(self.envs):
            _actions = action_index_to_actions(
                self._obs_space, action_spaces, actions, self.last_obs[i], i
            )
            obs = env.act_filter(_actions, obs_space)
            if obs.done:
                new_obs = env.reset_filter(obs_space)
                new_obs.done = True
                new_obs.reward = obs.reward
                new_obs.metrics = obs.metrics
                observations.append(new_obs)
            else:
                observations.append(obs)
        return self._batch_obs(observations)

    def _batch_obs(self, obs: List[Observation]) -> VecObs:
        self.last_obs = obs
        return batch_obs(obs, self.obs_space(), self.action_space())

    def __len__(self) -> int:
        return len(self.envs)

    def obs_space(self) -> ObsSpace:
        return self._obs_space

    def action_space(self) -> Dict[ActionName, ActionSpace]:
        return self._action_space


def action_index_to_actions(
    obs_space: ObsSpace,
    action_spaces: Dict[ActionName, ActionSpace],
    actions: Mapping[ActionName, RaggedBufferI64],
    last_obs: Observation,
    index: int = 0,
    probs: Optional[Dict[ActionName, npt.NDArray[np.float32]]] = None,
) -> Dict[ActionName, Action]:
    _actions: Dict[ActionName, Action] = {}
    for atype, action in actions.items():
        action_space = action_spaces[atype]
        if isinstance(action_space, GlobalCategoricalActionSpace):
            aindex = action[index].as_array()[0, 0]
            _actions[atype] = GlobalCategoricalAction(
                index=aindex,
                label=action_space.index_to_label[aindex],
                probs=probs[atype].reshape(-1) if probs is not None else None,
            )
            continue
        mask = last_obs.actions[atype]
        assert isinstance(mask, SelectEntityActionMask) or isinstance(
            mask, CategoricalActionMask
        )
        if mask.actor_ids is not None:
            actors = mask.actor_ids
        elif mask.actor_types is not None:
            actors = []
            for etype in mask.actor_types:
                actors.extend(last_obs.ids[etype])
        else:
            actors = []
            for ids in last_obs.ids.values():
                actors.extend(ids)
        aspace = action_spaces[atype]
        if isinstance(aspace, CategoricalActionSpace):
            _actions[atype] = CategoricalAction(
                actors=actors,
                indices=action[index].as_array().reshape(-1),
                index_to_label=aspace.index_to_label,
                probs=probs[atype] if probs is not None else None,
            )
        elif isinstance(action_spaces[atype], SelectEntityActionSpace):
            assert isinstance(mask, SelectEntityActionMask)
            if mask.actee_types is not None:
                index_to_actee: List[EntityID] = []
                for etype in mask.actee_types:
                    index_to_actee.extend(last_obs.ids[etype])
                actees = [
                    index_to_actee[a] for a in action[index].as_array().reshape(-1)
                ]
            elif mask.actee_ids is not None:
                actees = [
                    mask.actee_ids[e] for e in action[index].as_array().reshape(-1)
                ]
            else:
                index_to_id = last_obs.index_to_id(obs_space)
                actees = [index_to_id[e] for e in action[index].as_array().reshape(-1)]
            _actions[atype] = SelectEntityAction(
                actors=actors,
                actees=actees,
                probs=probs[atype] if probs is not None else None,
            )
        else:
            raise NotImplementedError(
                f"Action space type {type(action_spaces[atype])} not supported"
            )
    return _actions



import tempfile

import numpy as np
from ragged_buffer import RaggedBufferBool, RaggedBufferF32, RaggedBufferI64

from entity_gym.env import VecCategoricalActionMask, VecObs
from entity_gym.serialization import Sample, SampleRecorder, Trace


def test_serde_sample() -> None:
    sample = Sample(
        obs=VecObs(
            features={
                "hero": RaggedBufferF32.from_array(
                    np.array([[[1.0, 2.0, 0.3, 100.0, 10.0]]], dtype=np.float32),
                ),
                "enemy": RaggedBufferF32.from_array(
                    np.array(
                        [
                            [
                                [4.0, -2.0, 0.3, 100.0],
                                [5.0, -2.0, 0.3, 100.0],
                                [6.0, -2.0, 0.3, 100.0],
                            ]
                        ],
                        dtype=np.float32,
                    ),
                ),
                "box": RaggedBufferF32.from_array(
                    np.array(
                        [
                            [
                                [0.0, 0.0, 0.3, 100.0],
                                [1.0, 0.0, 0.3, 100.0],
                                [2.0, 0.0, 0.3, 100.0],
                            ]
                        ],
                        dtype=np.float32,
                    ),
                ),
            },
            visible={},
            action_masks={
                "move": VecCategoricalActionMask(
                    actors=RaggedBufferI64.from_array(np.array([[[0]]])),
                    mask=RaggedBufferBool.from_array(np.array([[[True, False, True]]])),
                ),
                "shoot": VecCategoricalActionMask(
                    actors=RaggedBufferI64.from_array(np.array([[[0]]])), mask=None
                ),
                "explode": VecCategoricalActionMask(
                    actors=RaggedBufferI64.from_array(np.array([[[4], [5], [6]]])),
                    mask=None,
                ),
            },
            reward=np.array([0.3124125987123489]),
            done=np.array([False]),
            metrics={},
        ),
        probs={
            "move": RaggedBufferF32.from_array(
                np.array([[[0.5], [0.2], [0.3], [0.0]]], dtype=np.float32)
            ),
            "shoot": RaggedBufferF32.from_array(
                np.array([[[0.9], [0.1]]], dtype=np.float32)
            ),
            "explode": RaggedBufferF32.from_array(
                np.array(
                    [[[0.3], [0.7]], [[0.2], [0.8]], [[0.1], [0.9]]], dtype=np.float32
                )
            ),
        },
        logits=None,
        actions={},
        step=[13],
        episode=[4213],
    )

    with tempfile.NamedTemporaryFile(mode="wb", delete=False) as f:
        sample_recorder = SampleRecorder(f.name, act_space=None, obs_space=None, subsample=1)  # type: ignore
        sample_recorder.record(sample)
        # modify the sample
        sample.obs.reward = np.array([1.0])
        sample.obs.features["hero"] = RaggedBufferF32.from_array(
            np.array([[[1.0, 2.0, 0.3, 200.0, 10.0]]], dtype=np.float32),
        )
        sample_recorder.record(sample)
        sample_recorder.close()

        with open(f.name, "rb") as f2:
            trace = Trace.deserialize(f2.read())
            assert len(trace.samples) == 2
            assert trace.samples[0].obs.reward[0] == 0.3124125987123489
            assert trace.samples[1].obs.reward[0] == 1.0
            assert (
                trace.samples[0].obs.action_masks["move"]
                == sample.obs.action_masks["move"]
            )
            np.testing.assert_equal(
                trace.samples[0].obs.features["hero"][0].as_array(),
                np.array([[1.0, 2.0, 0.3, 100.0, 10.0]], dtype=np.float32),
            )
            np.testing.assert_equal(
                trace.samples[1].obs.features["hero"][0].as_array(),
                np.array([[1.0, 2.0, 0.3, 200.0, 10.0]], dtype=np.float32),
            )



from dataclasses import dataclass
from typing import Dict, List, Optional, Set, Tuple

import msgpack_numpy
import numpy as np
import tqdm
from ragged_buffer import RaggedBufferBool, RaggedBufferF32, RaggedBufferI64

from entity_gym.env import ObsSpace
from entity_gym.env.environment import ActionSpace
from entity_gym.env.vec_env import VecActionMask
from entity_gym.ragged_dict import RaggedActionDict, RaggedBatchDict
from entity_gym.serialization.msgpack_ragged import ragged_buffer_decode
from entity_gym.serialization.sample_recorder import Sample


@dataclass
class Episode:
    number: int
    steps: int
    entities: Dict[str, RaggedBufferF32]
    visible: Dict[str, RaggedBufferBool]
    actions: Dict[str, RaggedBufferI64]
    masks: Dict[str, VecActionMask]
    logprobs: Dict[str, RaggedBufferF32]
    logits: Dict[str, RaggedBufferF32]
    total_reward: float
    complete: bool = False


@dataclass
class MergedSamples:
    entities: RaggedBatchDict[np.float32]
    visible: RaggedBatchDict[np.bool_]
    actions: RaggedBatchDict[np.int64]
    logprobs: RaggedBatchDict[np.float32]
    masks: RaggedActionDict
    logits: Optional[RaggedBatchDict[np.float32]]
    frames: int

    @classmethod
    def empty(clz) -> "MergedSamples":
        return MergedSamples(
            entities=RaggedBatchDict(RaggedBufferF32),
            visible=RaggedBatchDict(RaggedBufferBool),
            actions=RaggedBatchDict(RaggedBufferI64),
            logprobs=RaggedBatchDict(RaggedBufferF32),
            logits=None,
            masks=RaggedActionDict(),
            frames=0,
        )

    def push_sample(self, sample: Sample) -> None:
        self.entities.extend(sample.obs.features)
        self.visible.extend(sample.obs.visible)
        self.actions.extend(sample.actions)
        self.logprobs.extend(sample.probs)
        if sample.logits is not None:
            if self.logits is None:
                self.logits = RaggedBatchDict(RaggedBufferF32)
            self.logits.extend(sample.logits)
        self.masks.extend(sample.obs.action_masks)
        self.frames += len(sample.episode)


@dataclass
class Trace:
    action_space: Dict[str, ActionSpace]
    obs_space: ObsSpace
    samples: List[Sample]
    subsample: int = 1

    @classmethod
    def load(cls, path: str, progress_bar: bool = False) -> "Trace":
        with open(path, "rb") as f:
            return cls.deserialize(f.read(), progress_bar=progress_bar)

    @classmethod
    def deserialize(cls, data: bytes, progress_bar: bool = False) -> "Trace":
        samples: List[Sample] = []
        if progress_bar:
            pbar = tqdm.tqdm(total=len(data))

        offset = 0
        # Read version
        version = int(np.frombuffer(data[:8], dtype=np.uint64)[0])
        assert version == 0 or version == 1
        header_len = int(np.frombuffer(data[8:16], dtype=np.uint64)[0])
        header = msgpack_numpy.loads(
            data[16 : 16 + header_len],
            object_hook=ragged_buffer_decode,
            strict_map_key=False,
        )
        action_space = header["act_space"]
        obs_space = header["obs_space"]
        subsample = header.get("subsample", 1)

        offset = 16 + header_len
        while offset < len(data):
            size = int(np.frombuffer(data[offset : offset + 8], dtype=np.uint64)[0])
            offset += 8
            samples.append(Sample.deserialize(data[offset : offset + size]))
            offset += size
            if progress_bar:
                pbar.update(size + 8)
        return Trace(action_space, obs_space, samples, subsample=subsample)

    def episodes(
        self, include_incomplete: bool = False, progress_bar: bool = False
    ) -> List[Episode]:
        episodes = {}
        prev_episodes: Optional[List[int]] = None
        if progress_bar:
            samples = tqdm.tqdm(self.samples)
        else:
            samples = self.samples
        for sample in samples:
            for i, e in enumerate(sample.episode):
                if e not in episodes:
                    episodes[e] = Episode(
                        e,
                        0,
                        {},
                        {},
                        {},
                        {},
                        {},
                        {},
                        0.0,
                    )

                episodes[e].steps += 1
                episodes[e].total_reward += sample.obs.reward[i]
                if sample.obs.done[i] and prev_episodes is not None:
                    episodes[prev_episodes[i]].complete = True

                for name, feats in sample.obs.features.items():
                    if name not in episodes[e].entities:
                        episodes[e].entities[name] = feats[i]
                    else:
                        episodes[e].entities[name].extend(feats[i])
                for name, vis in sample.obs.visible.items():
                    if name not in episodes[e].visible:
                        episodes[e].visible[name] = vis[i]
                    else:
                        episodes[e].visible[name].extend(vis[i])
                for name, acts in sample.actions.items():
                    if name not in episodes[e].actions:
                        episodes[e].actions[name] = acts[i]
                    else:
                        episodes[e].actions[name].extend(acts[i])
                for name, mask in sample.obs.action_masks.items():
                    if name not in episodes[e].masks:
                        episodes[e].masks[name] = mask[i]
                    else:
                        episodes[e].masks[name].extend(mask[i])
                for name, logprobs in sample.probs.items():
                    if name not in episodes[e].logprobs:
                        episodes[e].logprobs[name] = logprobs[i]
                    else:
                        episodes[e].logprobs[name].extend(logprobs[i])
                if sample.logits is not None:
                    for name, logits in sample.logits.items():
                        if name not in episodes[e].logits:
                            episodes[e].logits[name] = logits[i]
                        else:
                            episodes[e].logits[name].extend(logits[i])
            prev_episodes = sample.episode
        return sorted(
            (e for e in episodes.values() if e.complete or include_incomplete),
            key=lambda e: e.number,
        )

    def train_test_split(
        self, test_frac: float = 0.1, progress_bar: bool = False
    ) -> Tuple[MergedSamples, MergedSamples]:
        if self.subsample == 1:
            total_frames = len(self.samples) * len(self.samples[0].episode)
        else:
            total_frames = sum(len(s.episode) for s in self.samples)
        if progress_bar:
            pbar = tqdm.tqdm(total=len(self.samples))

        test = MergedSamples.empty()
        test_episodes: Set[int] = set()
        i = 0
        while test.frames < total_frames * test_frac:
            sample = self.samples[i]
            test_episodes.update(sample.episode)
            test.push_sample(sample)
            i += 1
            if progress_bar:
                pbar.update(1)

        train = MergedSamples.empty()
        for sample in self.samples[i:]:
            # TODO: could be more efficient
            if any(e in test_episodes for e in sample.episode):
                continue
            train.push_sample(sample)
            if progress_bar:
                i += 1
                pbar.update(1)

        return train, test



from dataclasses import dataclass, field
from typing import Dict, Generic, Mapping, Type, TypeVar

import numpy as np
import numpy.typing as npt
from ragged_buffer import RaggedBuffer

from .env.vec_env import VecActionMask

ScalarType = TypeVar("ScalarType", bound=np.generic, covariant=True)


@dataclass
class RaggedBatchDict(Generic[ScalarType]):
    rb_cls: Type[RaggedBuffer[ScalarType]]
    buffers: Dict[str, RaggedBuffer[ScalarType]] = field(default_factory=dict)

    def extend(self, batch: Mapping[str, RaggedBuffer[ScalarType]]) -> None:
        for k, v in batch.items():
            if k not in self.buffers:
                self.buffers[k] = v
            else:
                self.buffers[k].extend(v)

    def clear(self) -> None:
        for buffer in self.buffers.values():
            buffer.clear()

    def __getitem__(
        self, index: npt.NDArray[np.int64]
    ) -> Dict[str, RaggedBuffer[ScalarType]]:
        return {k: v[index] for k, v in self.buffers.items()}


@dataclass
class RaggedActionDict:
    buffers: Dict[str, VecActionMask] = field(default_factory=dict)

    def extend(self, batch: Mapping[str, VecActionMask]) -> None:
        for k, v in batch.items():
            if k not in self.buffers:
                self.buffers[k] = v
            else:
                self.buffers[k].extend(v)

    def clear(self) -> None:
        for buffer in self.buffers.values():
            buffer.clear()

    def __getitem__(self, index: npt.NDArray[np.int64]) -> Dict[str, VecActionMask]:
        return {k: v[index] for k, v in self.buffers.items()}
